"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import scipy.sparse as sp

def aa_weights(Y, g, reg = 0, type = "lstsq", *args, **kwargs):
	""" Solve the constrained least-squares problem
		Minimize sum_squares(\\sum_{j=0}^m w_j * G^(k-m+j))
			subject to \\sum_{j=0}^m w_j = 1.
		with respect to w \\in \\reals^{m+1}.

	This can be transformed via a change of variables
		w_0 = c_0, w_j = c_j - c_{j-1} for j = 1,...,m-1, and w_m = 1 - c_{m-1}
	into the unconstrained problem
		Minimize sum_squares(g - Y*c)
	with respect to c \\in \\reals^m, where g_i = G^(i) and Y_k = [y_{k-m},...,y_{k-1}]
	for y_i = g_{i+1} - g_i.

	We add a regularization term for stability, so the final problem we solve is
		Minimize sum_squares(g - Y*c) + \\lambda*sum_squares(c)
	and return w as defined above.
	"""
	if type == "lstsq":
		if reg != 0:
			m = Y.shape[1]
			Y = np.vstack([Y, np.sqrt(reg)*np.eye(m)])
			g = np.concatenate([g, np.zeros(m)])
		gamma = np.linalg.lstsq(Y, g, *args, **kwargs)[0]
	elif type == "lsqr":
		if reg != 0:
			m = Y.shape[1]
			Y = sp.csc_matrix(Y)
			Y = sp.vstack([Y, np.sqrt(reg)*sp.eye(m)])
			g = np.concatenate([g, np.zeros(m)])
		gamma = sp.linalg.lsqr(Y, g, *args, **kwargs)[0]
	else:
		raise ValueError("Algorithm type not supported:", type)
	gamma_diff = np.diff(gamma, n=1)
	alpha = np.concatenate(([gamma[0]], gamma_diff, [1-gamma[-1]]))
	return alpha



import numpy as np
import warnings
from scipy import sparse
from scipy.special import expit, lambertw
from scipy.optimize import minimize
from a2dr.proximal.projection import proj_simplex
from a2dr.proximal.composition import prox_scale


def prox_kl(v, t = 1, u = None, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where 
    :math:`f(x) = \\sum_i x_i*\\log(x_i/u_i)`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(v):
        v = np.array([v])
    if u is None:
        u = np.ones(v.shape)
    return prox_scale(prox_kl_base, u, *args, **kwargs)(v, t)

def prox_logistic(v, t = 1, x0 = None, y = None, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where 
    :math:`f(x) = \\sum_i \\log(1 + \\exp(-y_i*x_i))`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(v):
        v = np.array([v])
    if x0 is None:
        # x0 = np.random.randn(*v.shape)
        x0 = v
    if y is None:
        y = -np.ones(v.shape)
    return prox_scale(prox_logistic_base, x0, y, *args, **kwargs)(v, t)

def prox_max(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = \\max_i x_i`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(v):
        v = np.array([v])
    return prox_scale(prox_max_base, *args, **kwargs)(v, t)

def prox_kl_base(v, t, u):
    """Proximal operator of :math:`f(x) = \\sum_i x_i\\log(x_i/u_i)`, where u is a given vector quantity.
    The function defaults to u_i = 1 for all i, so that :math:`f(x) = \\sum_i x_i\\log x_i`.
    """
    return t*lambertw(u*np.exp(v/t-1)/t)

def prox_logistic_base(v, t, x0, y):
    """Proximal operator of :math:`f(x) = \\sum_i \\log(1 + \\exp(-y_i*x_i))`, where y is a given vector quantity,
    solved using the Newton-CG method from scipy.optimize.minimize. The function defaults to y_i = -1 for all i,
    so that :math:`f(x) = \\sum_i \\log(1 + \\exp(x_i))`.
    """
    # Treat matrices elementwise.
    v_shape = v.shape
    v = v.flatten(order='C')
    x0 = x0.flatten(order='C')
    y = y.flatten(order='C')

    # Only works on dense vectors.
    if sparse.issparse(v):
        v = v.todense()
    if sparse.issparse(x0):
        x0 = x0.todense()
    if sparse.issparse(y):
        y = y.todense()

    # g(x) = \sum_i log(1 + exp(-y_i*x_i)) + 1/(2*t)*||x - v||_2^2
    def fun(x, y, v, t):
        # expit(x) = 1/(1 + exp(-x))
        return -np.sum(np.log(expit(np.multiply(y,x)))) + 1.0/(2*t)*np.sum((x - v)**2)

    # dg(x)/dx_i = -y_i/(1 + exp(y_i*x_i)) + (1/t)*(x_i - v_i)
    def jac(x, y, v, t):
        return -np.multiply(y, expit(-np.multiply(y,x))) + (1.0/t)*(x - v)

    # d^2g(x)/dx_i^2 = y_i^2*exp(y_i*x_i)/(1 + exp(y_i*x_i))^2 + 1/t
    def hess(x, y, v, t):
        return np.diag(np.multiply(np.multiply(y**2, np.exp(np.multiply(y,x))), expit(-np.multiply(y,x))**2) + 1.0/t)

    res = minimize(fun, x0, args=(y, v, t), method='Newton-CG', jac=jac, hess=hess)
    # res = minimize(fun, x0, args=(y, v, t), method='Newton-CG', jac=jac)
    if not res.success:
        warnings.warn(res.message)
    return res.x[0] if res.x.size == 1 else res.x.reshape(v_shape, order='C')

def prox_max_base(v, t):
    """Proximal operator of :math:`f(x) = \\max_i x_i`.
    """
    return v - t*proj_simplex(v/t)



import numpy as np
from scipy import sparse
from a2dr.proximal.interface import NUMPY_FUNS, SPARSE_FUNS
from a2dr.proximal.composition import prox_scale

def prox_box_constr(v, t = 1, v_lo = -np.inf, v_hi = np.inf, *args, **kwargs):
	"""Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f` is the set indicator that
	:math:`\\underline x \\leq x \\leq \\overline x`. Here the lower/upper bounds are (v_lo, v_hi), which default
	to (-Inf, Inf). The scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and 
	d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
	"""
	return prox_scale(prox_box_constr_base, v_lo, v_hi, *args, **kwargs)(v, t)

def prox_nonneg_constr(v, t = 1, *args, **kwargs):
	"""Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f` is the set indicator that
	:math:`x \\geq 0`. The scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and
	d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
	"""
	return prox_scale(prox_nonneg_constr_base, *args, **kwargs)(v, t)

def prox_nonpos_constr(v, t = 1, *args, **kwargs):
	"""Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f` is the set indicator that
	:math:`x \\leq 0`. The scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and
	d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
	"""
	return prox_scale(prox_nonpos_constr_base, *args, **kwargs)(v, t)

def prox_psd_cone(B, t = 1, *args, **kwargs):
	"""Proximal operator of :math:`tf(aB-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f` is the set indicator that
	:math:`B \\succeq 0` for :math:`B` a symmetric matrix. The scalar t > 0, and the optional arguments are
	a = scale, b = offset, c = lin_term, and d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default,
	t = 1, a = 1, b = 0, c = 0, and d = 0.
	"""
	if np.isscalar(B):
		B = np.array([[B]])
	if B.shape[0] != B.shape[1]:
		raise ValueError("B must be a square matrix.")
	B_symm = (B + B.T) / 2.0
	if not np.allclose(B, B_symm):
		raise ValueError("B must be a symmetric matrix.")
	return prox_scale(prox_psd_cone_base, *args, **kwargs)(B_symm, t)

def prox_soc(v, t = 1, *args, **kwargs):
	"""Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f` is the set indicator that
	:math:`\\|x_{1:n}\\|_2 \\leq x_{n+1}`. The scalar t > 0, and the optional arguments are a = scale, b = offset,
	c = lin_term, and d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, 	t = 1, a = 1, b = 0,
	c = 0, and d = 0.
	"""
	if np.isscalar(v) or not (len(v.shape) == 1 or (len(v.shape) == 2 and v.shape[1] == 1)):
		raise ValueError("v must be a vector")
	if v.shape[0] < 2:
		raise ValueError("v must have at least 2 elements.")
	return prox_scale(prox_soc_base, *args, **kwargs)(v, t)

def prox_box_constr_base(v, t, v_lo, v_hi):
	"""Proximal operator of the set indicator that :math:`\\underline x \\leq x \\leq \\overline x`.
	"""
	FUNS = SPARSE_FUNS if sparse.issparse(v) else NUMPY_FUNS
	max_elemwise, min_elemwise = FUNS["max_elemwise"], FUNS["min_elemwise"]
	return min_elemwise(max_elemwise(v, v_lo), v_hi)

def prox_nonneg_constr_base(v, t):
	"""Proximal operator of the set indicator that :math:`x \\geq 0`.
	"""
	# return prox_box_constr_base(v, t, 0, np.inf)
	# return v.maximum(0) if sparse.issparse(v) else np.maximum(v,0)
	FUNS = SPARSE_FUNS if sparse.issparse(v) else NUMPY_FUNS
	max_elemwise = FUNS["max_elemwise"]
	return max_elemwise(v, 0)

def prox_nonpos_constr_base(v, t):
	"""Proximal operator of the set indicator that :math:`x \\leq 0`.
	"""
	# return prox_box_constr_base(v, t, -np.inf, 0)
	# return v.minimum(0) if sparse.issparse(v) else np.minimum(v,0)
	FUNS = SPARSE_FUNS if sparse.issparse(v) else NUMPY_FUNS
	min_elemwise = FUNS["min_elemwise"]
	return min_elemwise(v, 0)

def prox_psd_cone_base(B, t):
	"""Proximal operator of the set indicator that :math:`B \\succeq 0`, where :math:`B` is a symmetric matrix.
	"""
	# B_symm = (B + B.T)/2.0
	# if not np.allclose(B, B_symm):
	#	raise ValueError("B must be a symmetric matrix.")
	# s, u = np.linalg.eigh(B_symm)
	s, u = np.linalg.eigh(B)
	s_new = np.maximum(s, 0)
	return u.dot(np.diag(s_new)).dot(u.T)

def prox_soc_base(v, t):
	"""Proximal operator of the set indicator that :math:`\\|v_{1:n}\\_2 \\leq v_{n+1}`, where :math:`v` is a vector of
	length n+1, `v_{1:n}` symbolizes its first n elements, and :math:`v_{n+1}` is its last element. This is equivalent
	to the projection of :math:`v` onto the second-order cone :math:`C = {(u,s):\\|u\\|_2 \\leq s}`.
	Parikh and Boyd (2013). "Proximal Algorithms." Foundations and Trends in Optimization. vol. 1, no. 3, Sect. 6.3.2.
	"""
	if sparse.issparse(v):
		v = v.tocsr()
		u = v[:-1]  # u = (v_1,...,v_n)
		s = v[-1]   # s = v_{n+1}
		s = np.asscalar(s.todense())

		u_norm = sparse.linalg.norm(u,'fro')
		if u_norm <= -s:
			return np.zeros(v.shape)
		elif u_norm <= s:
			return v
		else:
			scale = (1 + s / u_norm) / 2
			return scale * sparse.vstack((u, u_norm))
	else:
		u = v[:-1]  # u = (v_1,...,v_n)
		s = v[-1]   # s = v_{n+1}
		s = np.asscalar(s)

		u_norm = np.linalg.norm(u,2)
		if u_norm <= -s:
			return np.zeros(v.shape)
		elif u_norm <= s:
			return v
		else:
			scale = (1 + s / u_norm) / 2
			u_all = np.zeros(v.shape)
			u_all[:-1] = u
			u_all[-1] = u_norm
			return scale * u_all



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import numpy.linalg as LA
import scipy.sparse as sp
from scipy.stats.mstats import gmean
from time import time
from multiprocessing import Process, Pipe
import sys, os, warnings
from a2dr.precondition import precondition
from a2dr.acceleration import aa_weights
from a2dr.utilities import get_version

sys_stdout_origin = sys.stdout

def a2dr_worker(pipe, prox, v_init, A, t, anderson, m_accel):
    # Initialize AA-II parameters.
    if anderson:   # TODO: Store and update these efficiently as arrays.
        F_hist = []  # History of F(v^(k)).
    v_vec = v_init.copy()
    v_res = np.zeros(v_init.shape[0])

    # A2DR loop.
    while True:
        # Proximal step for x^(k+1/2).
        warnings.filterwarnings("ignore")
        sys.stdout = open(os.devnull, 'w')
        x_half = prox(v_vec, t)
        sys.stdout.close()
        sys.stdout = sys_stdout_origin
        warnings.filterwarnings("default")

        # Calculate v^(k+1/2) = 2*x^(k+1/2) - v^(k).
        v_half = 2*x_half - v_vec

        # Project to obtain x^(k+1) = v^(k+1/2) - A^T(AA^T)^{-1}(Av^(k+1/2) - b).
        pipe.send(v_half)
        dk, k = pipe.recv()   # dk = A^\dagger(Av^(k+1/2) - b)[i] for node i.
        x_new = v_half - dk

        if anderson and k > 0: # for k = 0, always do the vanilla DRS update
            m_k = min(m_accel, k)  # Keep F(v^(j)) for iterations (k-m_k) through k.

            # Save history of F(v^(k)).
            F_hist.append(v_vec + x_new - x_half)
            if len(F_hist) > m_k + 1:
                F_hist.pop(0)

            # Send s^(k-1) = v^(k) - v^(k-1) and g^(k) = v^(k) - F(v^(k)) = x^(k+1/2) - x^(k+1).
            pipe.send((v_res, x_half - x_new))

            # Receive safeguarding decision.
            AA_update = pipe.recv()
            if AA_update:
                # Receive AA-II weights for v^(k+1).
                alpha = pipe.recv()

                # Weighted update of v^(k+1).
                v_new = np.column_stack(F_hist).dot(alpha) #.dot(alpha[:(k + 1)]) ### Why truncate to (k+1)???
            else:
                # Revert to DRS update of v^(k+1).
                v_new = v_vec + x_new - x_half

            # Save v^(k+1) - v^(k) for next iteration.
            v_res = v_new - v_vec
        elif anderson and k == 0: 
            # Update v^(k+1) = v^(k) + x^(k+1) - x^(k+1/2).
            v_new = v_vec + x_new - x_half
            ## only useful when anderson = True but k == 0
            # Store v_res in case anderson = True
            v_res = v_new - v_vec
            # Update F_hist in case anderson = True
            F_hist.append(v_vec + x_new - x_half)
            # Send g^(k) = v^(k) - F(v^(k)) = x^(k+1/2) - x^(k+1).
            pipe.send(x_half - x_new)
        else:
            # Update v^(k+1) = v^(k) + x^(k+1) - x^(k+1/2).
            v_new = v_vec + x_new - x_half

        # Send x^(k+1/2) along with A*x^(k+1/2) and x^(k+1/2) - v^(k) for computing residuals.
        Ax_half = A.dot(x_half)
        xv_diff = x_half - v_vec
        pipe.send((x_half, Ax_half, xv_diff))
        
        v_vec = v_new


def a2dr(p_list, A_list = [], b = np.array([]), v_init = None, n_list = None, *args, **kwargs):
    start = time()

    # Problem parameters.
    max_iter = kwargs.pop("max_iter", 1000)
    t_init = kwargs.pop("t_init", 1/10)  # Step size.
    eps_abs = kwargs.pop("eps_abs", 1e-6)   # Absolute stopping tolerance.
    eps_rel = kwargs.pop("eps_rel", 1e-8)   # Relative stopping tolerance.
    precond = kwargs.pop("precond", True)  # Precondition A and b?
    ada_reg = kwargs.pop("ada_reg", True)   # Adaptive regularization?

    # AA-II parameters.
    anderson = kwargs.pop("anderson", True)
    m_accel = int(kwargs.pop("m_accel", 10))       # Maximum past iterations to keep (>= 0).
    lam_accel = kwargs.pop("lam_accel", 1e-8)      # AA-II regularization weight.
    aa_method = kwargs.pop("aa_method", "lstsq")   # Algorithm for solving AA LS problem.

    # Safeguarding parameters.
    D_safe = kwargs.pop("D_safe", 1e6)
    eps_safe = kwargs.pop("eps_safe", 1e-6)
    M_safe = kwargs.pop("M_safe", int(max_iter/100))

    # Printout parameters
    verbose = kwargs.pop("verbose", True)

    # Validate parameters.
    if max_iter <= 0:
        raise ValueError("max_iter must be a positive integer.")
    if t_init <= 0:
        raise ValueError("t_init must be a positive scalar.")
    if eps_abs < 0:
        raise ValueError("eps_abs must be a non-negative scalar.")
    if eps_rel < 0:
        raise ValueError("eps_rel must be a non-negative scalar.")
    if m_accel <= 0:
        raise ValueError("m_accel must be a positive integer.")
    if lam_accel < 0:
        raise ValueError("lam_accel must be a non-negative scalar.")
    if not aa_method in ["lstsq", "lsqr"]:
        raise ValueError("aa_method must be either 'lstsq' or 'lsqr'.")
    if D_safe < 0:
        raise ValueError("D_safe must be a non-negative scalar.")
    if eps_safe < 0:
        raise ValueError("eps_safe must be a non-negative scalar.")
    if M_safe <= 0:
        raise ValueError("M_safe must be a positive integer.")

    # DRS parameters.
    N = len(p_list)   # Number of subproblems.
    has_constr = len(A_list) != 0
    if len(A_list) == 0:
        if b.size != 0:
            raise ValueError("Dimension mismatch: nrow(A_i) != nrow(b)")
        if n_list is not None:
            if len(n_list) != N:
                raise ValueError("n_list must have exactly {} entries".format(N))
            A_list = [sp.csr_matrix((0, ni)) for ni in n_list]
        elif v_init is not None:
            if len(v_init) != N:
                raise ValueError("v_init must be None or contain exactly {} entries".format(N))
            A_list = [sp.csr_matrix((0, vi.shape[0])) for vi in v_init]
        else:
            raise ValueError("n_list or v_init must be defined if A_list and b are empty")
    if len(A_list) != N:
        raise ValueError("A_list must be empty or contain exactly {} entries".format(N))
    if v_init is None:
        # v_init = [np.random.randn(A.shape[1]) for A in A_list]
        v_init = [np.zeros(A.shape[1]) for A in A_list]
        # v_init = [sp.csc_matrix((A.shape[1],1)) for A in A_list]
    if len(v_init) != N:
        raise ValueError("v_init must be None or contain exactly {} entries".format(N))

    # Variable size list.
    if n_list is None:
        n_list = [A_list[i].shape[1] for i in range(N)]
    if len(n_list) != N:
        raise ValueError("n_list must be None or contain exactly {} entries".format(N))
    n_list_cumsum = np.insert(np.cumsum(n_list), 0, 0)

    for i in range(N):
        if A_list[i].shape[0] != b.shape[0]:
            raise ValueError("Dimension mismatch: nrow(A_i) != nrow(b)")
        elif A_list[i].shape[1] != v_init[i].shape[0]:
            raise ValueError("Dimension mismatch: ncol(A_i) != nrow(v_i)")
        elif A_list[i].shape[1] != n_list[i]:
            raise ValueError("Dimension mismatch: ncol(A_i) != n_i")
        if not sp.issparse(A_list[i]):
            A_list[i] = sp.csr_matrix(A_list[i])

    if verbose:
        version = get_version("__init__.py")
        line_solver = "a2dr v" + version + " - Prox-Affine Distributed Convex Optimization Solver"
        dashes = "-" * len(line_solver)
        ddashes = "=" * len(line_solver)
        line_authors = "(c) Anqi Fu, Junzi Zhang"
        num_spaces_authors = (len(line_solver) - len(line_authors)) // 2
        line_affil = "Stanford University   2019"
        num_spaces_affil = (len(line_solver) - len(line_affil)) // 2
        print(dashes)
        print(line_solver)
        print(" " * num_spaces_authors + line_authors)
        print(" " * num_spaces_affil + line_affil)
        print(dashes)

    # Precondition data.
    if precond and has_constr:
        if verbose:
            print('### Preconditioning starts ... ###')
        p_list, A_list, b, e_pre = precondition(p_list, A_list, b)
        t_init = 1/gmean(e_pre)**2/10
        if verbose:
            print('### Preconditioning finished.  ###')

    if verbose:
        print("max_iter = {}, t_init (after preconditioning) = {:.2f}".format(
               max_iter, t_init))
        print("eps_abs = {:.2e}, eps_rel = {:.2e}, precond = {!r}".format(
               eps_abs, eps_rel, precond))
        print("ada_reg = {!r}, anderson = {!r}, m_accel = {}".format(
               ada_reg, anderson, m_accel))
        print("lam_accel = {:.2e}, aa_method = {}, D_safe = {:.2e}".format(
               lam_accel, aa_method, D_safe))
        print("eps_safe = {:.2e}, M_safe = {:d}".format(
               eps_safe, M_safe))

    # Store constraint matrix for projection step.
    A = sp.csr_matrix(sp.hstack(A_list))
    if verbose:
        print("variables n = {}, constraints m = {}".format(A.shape[1], A.shape[0]))
        print("nnz(A) = {}".format(A.nnz))
        print("Setup time: {:.2e}".format(time() - start))

    # Check linear feasibility
    sys.stdout = open(os.devnull, 'w')
    r1norm = sp.linalg.lsqr(A, b)[3]
    sys.stdout.close()
    sys.stdout = sys_stdout_origin
    if r1norm >= np.sqrt(eps_abs): # infeasible
        if verbose:
            print('Infeasible linear equality constraint: minimum constraint violation = {:.2e}'.format(r1norm))
            print('Status: Terminated due to linear infeasibility')
            print("Solve time: {:.2e}".format(time() - start))
        return {"x_vals": None, "primal": None, "dual": None, "num_iters": None, "solve_time": None}

    if verbose:
        print("----------------------------------------------------")
        print(" iter | total res | primal res | dual res | time (s)")
        print("----------------------------------------------------")

    # Set up the workers.
    pipes = []
    procs = []
    for i in range(N):
        local, remote = Pipe()
        pipes += [local]
        procs += [Process(target=a2dr_worker, args=(remote, p_list[i], v_init[i], A_list[i], \
                                                    t_init, anderson, m_accel) + args)]
        procs[-1].start()

    # Initialize AA-II variables.
    if anderson:   # TODO: Store and update these efficiently as arrays.
        n_sum = np.sum([np.prod(v.shape) for v in v_init])
        g_vec = np.zeros(n_sum)   # g^(k) = v^(k) - F(v^(k)).
        s_hist = []  # History of s^(j) = v^(j+1) - v^(j), kept in S^(k) = [s^(k-m_k) ... s^(k-1)].
        y_hist = []  # History of y^(j) = g^(j+1) - g^(j), kept in Y^(k) = [y^(k-m_k) ... y^(k-1)].
        n_AA = M_AA = 0   # Safeguarding counters.

    # A2DR loop.
    k = 0
    finished = False
    safeguard = True
    r_primal = np.zeros(max_iter)
    r_dual = np.zeros(max_iter)
    r_best = np.inf

    # Warm start terms.
    dk = np.zeros(A.shape[1])
    sol = np.zeros(A.shape[0])

    while not finished:
        # Gather v_i^(k+1/2) from nodes.
        v_halves = [pipe.recv() for pipe in pipes]

        # Projection step for x^(k+1).
        v_half = np.concatenate(v_halves, axis=0)
        sys.stdout = open(os.devnull, 'w')
        dk = sp.linalg.lsqr(A, A.dot(v_half) - b, atol=1e-10, btol=1e-10, x0=dk)[0]
        sys.stdout.close()
        sys.stdout = sys_stdout_origin

        # Scatter d^k = A^\dagger(Av^(k+1/2) - b).
        for i in range(N):
            pipes[i].send((dk[n_list_cumsum[i]:n_list_cumsum[i+1]], k))

        if anderson and k > 0: # for k = 0, always do the vanilla DRS update
            m_k = min(m_accel, k)  # Keep (y^(j), s^(j)) for iterations (k-m_k) through (k-1).

            # Gather s_i^(k-1) and g_i^(k) from nodes.
            sg_update = [pipe.recv() for pipe in pipes]
            s_new, g_new = map(list, zip(*sg_update))
            s_new = np.concatenate(s_new, axis=0)   # s_i^(k-1) = v_i^(k) - v_i^(k-1).
            g_new = np.concatenate(g_new, axis=0)   # g_i^(k) = v_i^(k) - F(v_i^(k)) = x_i^(k+1/2) - x_i^(k+1).

            # Save newest column y^(k-1) = g^(k) - g^(k-1) of matrix Y^(k).
            y_hist.append(g_new - g_vec)
            if len(y_hist) > m_k:
                y_hist.pop(0)
            g_vec = g_new

            # Save newest column s^(k-1) = v^(k) - v^(k-1) of matrix S^(k).
            s_hist.append(s_new)
            if len(s_hist) > m_k:
                s_hist.pop(0)

            # Safeguard update.
            if safeguard or M_AA >= M_safe:
                if LA.norm(g_vec) <= D_safe*g0_norm*(n_AA/M_safe + 1)**(-(1 + eps_safe)):
                    AA_update = True
                    n_AA = n_AA + 1
                    M_AA = 0
                    safeguard = False
                else:
                    AA_update = False
                    M_AA = 0
            else:
                AA_update = True
                M_AA = M_AA + 1
                n_AA = n_AA + 1

            # Scatter safeguarding decision.
            for pipe in pipes:
                pipe.send(AA_update)
            if AA_update:
                # Compute and scatter AA-II weights.
                Y_mat = np.column_stack(y_hist)
                S_mat = np.column_stack(s_hist)
                if ada_reg:
                    reg = lam_accel * (LA.norm(Y_mat)**2 + LA.norm(S_mat)**2)  # AA-II regularization.
                else:
                    reg = lam_accel
                alpha = aa_weights(Y_mat, g_new, reg, type=aa_method, rcond=None)
                for pipe in pipes:
                    pipe.send(alpha)
              
        elif anderson and k == 0:
            AA_update = False   # Initial step is always DRS.
            g_new = [pipe.recv() for pipe in pipes]
            g_vec = np.concatenate(g_new, axis=0)
            g0_norm = LA.norm(g_vec)

        # Compute l2-norm of primal and dual residuals.
        r_update = [pipe.recv() for pipe in pipes]
        x_halves, Ax_halves, xv_diffs = map(list, zip(*r_update))
        r_primal_vec = sum(Ax_halves) - b
        r_primal[k] = LA.norm(r_primal_vec, ord=2)

        subgrad = np.concatenate(xv_diffs)/t_init
        # sol = LA.lstsq(A.T, subgrad, rcond=None)[0]
        sys.stdout = open(os.devnull, 'w')
        sol = sp.linalg.lsqr(A.T, subgrad, atol=1e-10, btol=1e-10, x0=sol)[0]
        sys.stdout.close()
        sys.stdout = sys_stdout_origin
        r_dual_vec = A.T.dot(sol) - subgrad
        r_dual[k] = LA.norm(r_dual_vec, ord=2)

        # Save x_i^(k+1/2) if residual norm is smallest so far.
        r_all = LA.norm(np.concatenate([r_primal_vec, r_dual_vec]), ord=2)
        if k == 0:   # Store ||r^0||_2 for stopping criterion.
            r_all_0 = r_all
        if k == 0 or r_all < r_best:
            x_final = x_halves
            r_best = r_all
            k_best = k

        if (k % 100 == 0 or k == max_iter-1)and verbose:
            # print every 100 iterations or reaching maximum
            print("{}| {}  {}  {}  {}".format(str(k).rjust(6), 
                                        format(r_all, ".2e").ljust(10),
                                        format(r_primal[k], ".2e").ljust(11), 
                                        format(r_dual[k], ".2e").ljust(9),
                                        format(time() - start, ".2e").ljust(8)))

        # Stop when residual norm falls below tolerance.
        k = k + 1
        finished = k >= max_iter or (r_all <= eps_abs + eps_rel * r_all_0)
        if r_all <= eps_abs + eps_rel * r_all_0 and k % 100 != 0:
            # print the best iterate
            print("{}| {}  {}  {}  {}".format(str(k-1).rjust(6), 
                            format(r_all, ".2e").ljust(10),
                            format(r_primal[k-1], ".2e").ljust(11), 
                            format(r_dual[k-1], ".2e").ljust(9),
                            format(time() - start, ".2e").ljust(8)))

    # Unscale and return x_i^(k+1/2).
    [p.terminate() for p in procs]
    if precond and has_constr:
        x_final = [ei*x for x, ei in zip(x_final, e_pre)]
    end = time()
    if verbose:
        print("----------------------------------------------------")
        if k < max_iter:
            print("Status: Solved")
        else:
            print("Status: Reach maximum iterations")
        print("Solve time: {:.2e}".format(end - start))
        print("Total number of iterations: {}".format(k))
        print("Best total residual: {:.2e}; reached at iteration {}".format(r_best, k_best))
        print(ddashes)
    return {"x_vals": x_final, "primal": np.array(r_primal[:k]), "dual": np.array(r_dual[:k]), \
            "num_iters": k, "solve_time": (end - start)}



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""

from a2dr.proximal.constraint import prox_box_constr, prox_nonneg_constr, prox_nonpos_constr, prox_psd_cone, prox_soc
from a2dr.proximal.elementwise import prox_abs, prox_constant, prox_exp, prox_huber, prox_identity, prox_neg, \
                                      prox_neg_entr, prox_neg_log, prox_pos
from a2dr.proximal.matrix import prox_neg_log_det, prox_sigma_max, prox_trace
from a2dr.proximal.misc import prox_kl, prox_logistic, prox_max
from a2dr.proximal.norm import prox_norm1, prox_norm2, prox_norm_inf, prox_norm_fro, prox_norm_nuc, prox_group_lasso
from a2dr.proximal.quadratic import prox_qp, prox_quad_form, prox_sum_squares, prox_sum_squares_affine



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""

# Base class for unit tests.
from unittest import TestCase
import numpy as np
import matplotlib.pyplot as plt

class BaseTest(TestCase):
    # AssertAlmostEqual for lists.
    def assertItemsAlmostEqual(self, a, b, places=4):
        if np.isscalar(a):
            a = [a]
        else:
            a = self.mat_to_list(a)
        if np.isscalar(b):
            b = [b]
        else:
            b = self.mat_to_list(b)
        for i in range(len(a)):
            self.assertAlmostEqual(a[i], b[i], places)

    # Overridden method to assume lower accuracy.
    def assertAlmostEqual(self, a, b, places=4):
        super(BaseTest, self).assertAlmostEqual(a.real, b.real, places=places)
        super(BaseTest, self).assertAlmostEqual(a.imag, b.imag, places=places)

    def mat_to_list(self, mat):
        """Convert a numpy matrix to a list.
        """
        if isinstance(mat, (np.matrix, np.ndarray)):
            return np.asarray(mat).flatten('F').tolist()
        else:
            return mat

    def plot_residuals(self, r_primal, r_dual, normalize = False, show = True, title = None, semilogy = False, savefig = None, *args, **kwargs):
        if normalize:
            r_primal = r_primal / r_primal[0] if r_primal[0] != 0 else r_primal
            r_dual = r_dual / r_dual[0] if r_dual[0] != 0 else r_dual

        if semilogy:
            plt.semilogy(range(len(r_primal)), r_primal, label = "Primal", *args, **kwargs)
            plt.semilogy(range(len(r_dual)), r_dual, label = "Dual", *args, **kwargs)
        else:
            plt.plot(range(len(r_primal)), r_primal, label = "Primal", *args, **kwargs)
            plt.plot(range(len(r_dual)), r_dual, label = "Dual", *args, **kwargs)
        plt.legend()
        plt.xlabel("Iteration")
        plt.ylabel("Residual")
        if title:
            plt.title(title)
        if show:
            plt.show()
        if savefig:
            plt.savefig(savefig, bbox_inches="tight")

    def compare_results(self, probs, obj_a2dr, obj_comb, x_a2dr, x_comb):
        N = len(probs.variables())
        for i in range(N):
            print("\nA2DR Solution:\n", x_a2dr[i])
            print("Base Solution:\n", x_comb[i])
            print("MSE: ", np.mean(np.square(x_a2dr[i] - x_comb[i])), "\n")
        print("A2DR Objective: %f" % obj_a2dr)
        print("Base Objective: %f" % obj_comb)
        print("Iterations: %d" % probs.solver_stats["num_iters"])
        print("Elapsed Time: %f" % probs.solver_stats["solve_time"])

    def compare_residuals(self, res_drs, res_a2dr, m_vals):
        if not isinstance(res_a2dr, list):
            res_a2dr = [res_a2dr]
        if not isinstance(m_vals, list):
            m_vals = [m_vals]
        if len(m_vals) != len(res_a2dr):
            raise ValueError("Must have same number of AA-II residuals as memory parameter values")

        plt.semilogy(range(res_drs.shape[0]), res_drs, label="DRS")
        for i in range(len(m_vals)):
            label = "A2DR (m = {})".format(m_vals[i])
            plt.semilogy(range(res_a2dr[i].shape[0]), res_a2dr[i], linestyle="--", label=label)
        plt.legend()
        plt.xlabel("Iteration")
        plt.ylabel("Residual")
        plt.show()

    def compare_primal_dual(self, drs_result, a2dr_result, savefig = None):
        # Compare residuals
        plt.semilogy(range(drs_result["num_iters"]), drs_result["primal"], color="blue", linestyle="--",
                     label="Primal (DRS)")
        plt.semilogy(range(a2dr_result["num_iters"]), a2dr_result["primal"], color="blue", label="Primal (A2DR)")
        plt.semilogy(range(drs_result["num_iters"]), drs_result["dual"], color="darkorange", linestyle="--",
                     label="Dual (DRS)")
        plt.semilogy(range(a2dr_result["num_iters"]), a2dr_result["dual"], color="darkorange", label="Dual (A2DR) ")
        # plt.title("Residuals")
        plt.legend()
        if savefig:
            plt.savefig(savefig, bbox_inches="tight")
        plt.show()
        
    def compare_total(self, drs_result, a2dr_result, savefig = None):
        # Compare residuals
        plt.semilogy(range(drs_result["num_iters"]), np.sqrt(drs_result["primal"]**2+drs_result["dual"]**2), color="blue", label="Residuals (DRS)")
        plt.semilogy(range(a2dr_result["num_iters"]), np.sqrt(a2dr_result["primal"]**2+a2dr_result["dual"]**2), color="darkorange", label="Residuals (A2DR)")
        # plt.title("Residuals")
        plt.legend()
        if savefig:
            plt.savefig(savefig, bbox_inches="tight")
        plt.show()
        
    def compare_total_all(self, results, names, savefig = None):
        # Compare residuals in the results list
        # len(names) must be equal to len(results)
        for i in range(len(names)):
            result = results[i]
            name = names[i]
            plt.semilogy(range(result["num_iters"]), np.sqrt(result["primal"]**2+result["dual"]**2), 
                         label="Residuals (" + name + ")")
        # plt.title("Residuals")
        plt.legend()
        if savefig:
            plt.savefig(savefig, bbox_inches="tight")
        plt.show()



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import codecs
import os.path

# code for single sourcing versions
# reference: https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

def solve_vec(x, var_type='vec'):
	# reshape to vector format for input to a2dr
	if var_type == 'vec':
		return x, x.shape
	elif var_type == 'mat_C':
		return x.ravel(order='C'), x.shape
	elif var_type == 'mat_F':
		return x.ravel(order='F'), x.shape
	elif var_type == 'mat_symm':
		# lower triangular part, row-wise stacking
		if x.shape[0] != x.shape[1] or np.linalg.norm(x-x.T) != 0:
			raise ValueError("input must be square and symmetric")
		mask = np.ones(Q.shape, dtype=bool)
		np.fill_diagonal(mask, 0)
		x[mask] *= np.sqrt(2)
		ind = np.tril_indices(x.shape[0])
		return x[ind], x.shape
	else:
		raise ValueError("var_type = must be vec, mat_C, mat_F or mat_symm")

def solve_mat(x, shape=None, var_type='vec'):
	# reshape back to the original format after running a2dr
	if var_type == 'vec':
		return x
	elif var_type == 'mat_C':
		if shape == None:
			raise ValueError("shape must be provided for var_type = mat_C")
		return x.reshape(shape, order='C')
	elif var_type == 'mat_F':
		if shape == None:
			raise ValueError("shape must be provided for var_type = mat_F")
		return x.reshape(shape, order='F')
	elif var_type == 'mat_symm':
		# lower triangular part, row-wise stacking
		if x.shape[0] != x.shape[1] or np.linalg.norm(x-x.T) != 0:
			raise ValueError("input must be square and symmetric")
		if shape == None:
			raise ValueError("shape must be provided for var_type = mat_symm")
		ind = np.tril_indices(shape[0])
		ind_u = np.triu_indices(shape[0])
		y = np.zeros(shape)
		y[ind] = x
		y[ind_u] = y.T[ind_u]
		return y
	else:
		raise ValueError("var_type = must be vec, mat_C, mat_F or mat_symm")



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy import sparse
from scipy.sparse import block_diag, issparse, csr_matrix, csc_matrix, diags
from scipy.sparse.linalg import norm
import scipy.linalg as sLA
from scipy.stats.mstats import gmean

def precondition(p_list, A_list, b, tol = 1e-3, max_iter = 1000):
    # print('### Preconditioning starts ...')
    if all([Ai.size == 0 for Ai in A_list]):
        return p_list, A_list, b, np.ones(len(A_list))

    n_list = [A.shape[1] for A in A_list]
    sparse_check = [sparse.issparse(A) for A in A_list]
    # Enforce csr format for better matrix operation efficiency.
    if np.sum(sparse_check) == 0:      # all dense
        A = csr_matrix(np.hstack(A_list))
    elif np.prod(sparse_check) == 1:   # all sparse
        A = csr_matrix(sparse.hstack(A_list))
    else:
        A_list_csr = [csr_matrix(A) for A in A_list]
        A = csr_matrix(sparse.hstack(A_list))
    d, e, A_hat, k = mat_equil(A, n_list, tol, max_iter)

    split_idx = np.cumsum(n_list)
    split_idx = np.hstack([0, split_idx])
    A_hat = csc_matrix(A_hat)   # faster column slicing
    A_eq_list = [A_hat[:,split_idx[i]:split_idx[i+1]] for i in range(len(n_list))]
    A_eq_list = [csr_matrix(A_eq_list[i]) for i in range(len(A_eq_list))]   # change back to csr format

    # Note: We must do it this way because the standard pythonic list comprehension, i.e., [f(x) for x in iterable]
    # will create *duplicate* function handles, leading to incorrect results! This is due to late binding:
    # https://stackoverflow.com/questions/3431676/creating-functions-in-a-loop
    # https://docs.python-guide.org/writing/gotchas/#late-binding-closures
    def proto(i, p_list, e):
        return lambda v, t: p_list[i](e[i]*v, t*e[i]**2)/e[i]
    p_eq_list = list(map(lambda i: proto(i,p_list,e), range(len(p_list))))
    return p_eq_list, A_eq_list, d*b, e

def mat_equil(A, n_list, tol, max_iter):
    # Block matrix equilibration using regularized Sinkhorn-Knopp
    # Reference: [POGS] http://stanford.edu/~boyd/papers/pdf/pogs.pdf
    '''
    1. Input
    A: a numpy 2d-array or matrix of size m-by-n to be equilibrated
    n_list: a list of size N containing size n_i of each variable block i (n_1+...+n_N = n)
    tol: tolerance for termination
    max_iter: maximum number of iterations
   
    2. Output
    d: left scaling vector (of size m)
    e: right scaling vector (of size N)
    B: equilibrated matrix B = diag(d) A diag(e_1I_{n_1},...,e_NI_{n_N})
    k: number of iterations terminated
    
    3. Requirement:
    import numpy as np
    from scipy.linalg import block_diag
    '''
    
    N = len(n_list)
    m = A.shape[0]
    em, eN = np.ones(m), np.ones(N)
    d, e = em, eN
    
    # Form the size m-by-N matrix A_block, whose (i,j)-th entry is \sum_{k=n_1+...+n_{j-1}}}^{n_1+...+n_j} A_{ik}^2
    gamma = (m + N)/m/N * np.sqrt(np.finfo(float).eps)
    A2 = A.power(2) if issparse(A) else np.power(A,2)
    ave_list = [np.ones([n_i,1]) for n_i in n_list]
    A_block = A2.dot(block_diag(ave_list))
    A_block_T = A_block.transpose()
    # print('Block matrix shape = {}'.format(A_block.shape))
    # print('gamma={}'.format(gamma))
    
    # Apply regularized Sinkhorn-Knopp on A_block
    for k in range(max_iter):
        d1 = N / (A_block.dot(e) + N * gamma * em)
        e1 = m / (A_block_T.dot(d1) + m * gamma * eN)
        err_d = np.linalg.norm(d1 - d)
        err_e = np.linalg.norm(e1 - e)
        d = d1
        e = e1
        # print('k={}, err_d={}, err_e={}'.format(k, err_d/np.sqrt(m), err_e/np.sqrt(N)))
        if err_d/np.sqrt(m) <= tol and err_e/np.sqrt(N) <= tol:
            break
    
    d = np.sqrt(d)
    e = np.sqrt(e)
    I_list = [sparse.eye(n_list[i]) * e[i] for i in range(N)]
    E = csr_matrix(block_diag(I_list))
    D = csr_matrix(diags(d))
    # print('generate D, E')
    B = D.dot(csr_matrix(A).dot(E))
    # print('compute scaled matrix')
    
    # Rescale to have \|DAE\|_2 close to 1
    scale = norm(B, 'fro') / np.sqrt(np.min([m,N]))
    d_mean = gmean(d)
    e_mean = gmean(e)
    q = np.log(d_mean / e_mean * scale) / 2 / np.log(scale)
    d = d * (scale ** (-q))
    e = e * (scale ** (q-1))
    B = B / scale
    
    return d, e, B, k



import numpy as np
from scipy import sparse
from a2dr.proximal.composition import prox_scale
from a2dr.proximal.norm import prox_norm_inf_base

def prox_neg_log_det(B, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(aB-b) + cB + d\\|B\\|_F^2`, where :math:`f(B) = -\\log\\det(B)`, where `B` is a
    symmetric matrix. The scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and
    d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if B.shape[0] != B.shape[1]:
        raise ValueError("B must be a square matrix.")
    B_symm = (B + B.T) / 2.0
    if not (np.allclose(B, B_symm)):
        raise ValueError("B must be a symmetric matrix.")
    return prox_scale(prox_neg_log_det_base, *args, **kwargs)(B_symm, t)

def prox_sigma_max(B, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(aB-b) + cB + d\\|B\\|_F^2`, where :math:`f(B) = \\sigma_{\\max}(B)`
    is the maximum singular value of :math:`B`, for scalar t > 0, and the optional arguments are a = scale,
    b = offset, c = lin_term, and d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default,
    t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_sigma_max_base, *args, **kwargs)(B, t)

def prox_trace(B, t = 1, C = None, *args, **kwargs):
    """Proximal operator of :math:`tf(aB-b) + cB + d\\|B\\|_F^2`, where :math:`f(B) = tr(C^TB)` is the trace of
    :math:`C^TB`, where C is a given matrix quantity. By default, C is the identity matrix, so :math:`f(B) = tr(B)`.
    The scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(B):
        B = np.array([[B]])
    if C is None:
        C = sparse.eye(B.shape[0])
    if B.shape[0] != C.shape[0]:
        raise ValueError("Dimension mismatch: nrow(B) != nrow(C)")
    if B.shape[1] != C.shape[1]:
        raise ValueError("Dimension mismatch: ncol(B) != ncol(C)")
    return prox_scale(prox_trace_base, C=C, *args, **kwargs)(B, t)

def prox_neg_log_det_base(B, t):
    """Proximal operator of :math:`f(B) = -\\log\\det(B)`.
    """
    s, u = np.linalg.eigh(B)
    # s_new = (s + np.sqrt(s**2 + 4*t))/2
    id_pos = (s >= 0)
    id_neg = (s < 0)
    s_new = np.zeros(len(s))
    s_new[id_pos] = (s[id_pos] + np.sqrt(s[id_pos] ** 2 + 4.0 * t)) / 2
    s_new[id_neg] = 2.0 * t / (np.sqrt(s[id_neg] ** 2 + 4.0 * t) - s[id_neg])
    return u.dot(np.diag(s_new)).dot(u.T)

def prox_sigma_max_base(B, t):
    """Proximal operator of :math:`f(B) = \\sigma_{\\max}(B)`, the maximum singular value of :math:`B`, otherwise
    known as the spectral norm.
    """
    U, s, Vt = np.linalg.svd(B, full_matrices=False)
    s_new = prox_norm_inf_base(s, t)
    return U.dot(np.diag(s_new)).dot(Vt)

def prox_trace_base(B, t, C):
    """Proximal operator of :math:`f(B) = tr(C^TB)`, the trace of :math:`C^TB`, where C is a given matrix quantity
    such that :math:`C^TB` is square.
    """
    return B - t*C



import numpy as np
import scipy as sp
from scipy import sparse

def shape_to_2d(shape):
    if np.isscalar(shape) or len(shape) == 0:
        return 1,1
    elif len(shape) == 1:
        return shape[0],1
    else:
        return shape

def apply_to_nonzeros(fun, v):
    if sparse.issparse(v):
        v_new = v.copy()
        v_new.data = fun(v.data)
        return v_new
    else:
        return fun(v)

NUMPY_FUNS = {"hstack": np.hstack,
              "max_elemwise": np.maximum,
              "min_elemwise": np.minimum,
              "mul_elemwise": np.multiply,
              "norm": np.linalg.norm,
              "vstack": np.vstack,
              "zeros": lambda shape, dtype=float: np.zeros(shape, dtype=dtype)
              }

SPARSE_FUNS = {"hstack": sparse.hstack,
               "max_elemwise": lambda x, y: x.maximum(y),
               "min_elemwise": lambda x, y: x.minimum(y),
               "mul_elemwise": lambda x, y: x.multiply(y),
               "norm": sp.sparse.linalg.norm,
               "vstack": sparse.vstack,
               "zeros": lambda shape, dtype=None: sparse.csr_matrix(shape_to_2d(shape), dtype=dtype)
               }



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import numpy.linalg as LA
from scipy import sparse

from a2dr import a2dr
from a2dr.proximal import prox_sum_squares_affine, prox_nonneg_constr
from a2dr.tests.base_test import BaseTest

class TestBasic(BaseTest):
    """Unit tests for A2DR paper experiments."""

    def setUp(self):
        np.random.seed(1)
        self.eps_rel = 1e-8  # specify these in all examples?
        self.eps_abs = 1e-6
        self.MAX_ITER = 1000

    def test_unconstrained(self):
        # minimize ||y - X\beta||_2^2.

        # Problem data.
        m, n = 100, 80
        density = 0.1
        X = sparse.random(m, n, density=density, data_rvs=np.random.randn)
        y = np.random.randn(m)
        prox_list = [lambda v, t: prox_sum_squares_affine(v, t, F=X, g=y, method="lstsq")]

        # Solve with NumPy.
        np_result = LA.lstsq(X.todense(), y, rcond=None)
        np_beta = np_result[0]
        np_obj = np.sum(np_result[1])

        # Solve with DRS.
        drs_result = a2dr(prox_list, n_list=[n], anderson=False, max_iter=self.MAX_ITER)
        drs_beta = drs_result["x_vals"][-1]
        drs_obj = np.sum((y - X.dot(drs_beta))**2)
        print("Finish DRS.")

        # Solve with A2DR.
        a2dr_result = a2dr(prox_list, n_list=[n], anderson=True, max_iter=self.MAX_ITER)
        a2dr_beta = a2dr_result["x_vals"][-1]
        a2dr_obj = np.sum((y - X.dot(a2dr_beta))**2)
        print("Finish A2DR.")

        self.assertAlmostEqual(np_obj, drs_obj)
        self.assertAlmostEqual(np_obj, a2dr_obj)

    def test_ols(self):
        # minimize ||y - X\beta||_2^2.
        m = 100
        n = 10
        N = 4  # Number of splits. (split X row-wise)
        beta_true = np.array(np.arange(-n / 2, n / 2) + 1)
        X = np.random.randn(m, n)
        y = X.dot(beta_true) + np.random.randn(m)

        # Split problem.
        X_split = np.split(X, N)
        y_split = np.split(y, N)

        # Construct list of proximal operators.
        # Note: We must do it this way to avoid problems caused by late binding:
        # https://docs.python-guide.org/writing/gotchas/#late-binding-closures
        prox_list = [lambda v, t, i=i: prox_sum_squares_affine(v, t, F=X_split[i], g=y_split[i], method="lstsq") \
                        for i in range(N)]
        v_init = N * [np.random.randn(n)]

        # Solve with NumPy.
        np_beta = []
        np_obj = 0
        for i in range(N):
            np_result = LA.lstsq(X_split[i], y_split[i], rcond=None)
            np_beta += [np_result[0]]
            np_obj += np.sum(np_result[1])
        print("NumPy Objective:", np_obj)
        print("NumPy Solution:", np_beta)

        # Solve with DRS (proximal point method).
        drs_result = a2dr(prox_list, v_init=v_init, max_iter=self.MAX_ITER, eps_abs=self.eps_abs, \
                          eps_rel=self.eps_rel, anderson=False)
        drs_beta = drs_result["x_vals"]
        drs_obj = np.sum([(yi - Xi.dot(beta)) ** 2 for yi, Xi, beta in zip(y_split, X_split, drs_beta)])
        print("DRS Objective:", drs_obj)
        print("DRS Solution:", drs_beta)

        # Solve with A2DR (proximal point method with Anderson acceleration).
        a2dr_result = a2dr(prox_list, v_init=v_init, max_iter=self.MAX_ITER, eps_abs=self.eps_abs, \
                           eps_rel=self.eps_rel, anderson=True)
        a2dr_beta = a2dr_result["x_vals"]
        a2dr_obj = np.sum([(yi - Xi.dot(beta)) ** 2 for yi, Xi, beta in zip(y_split, X_split, drs_beta)])
        print("A2DR Objective:", a2dr_obj)
        print("A2DR Solution:", a2dr_beta)

        # Compare results.
        self.assertAlmostEqual(np_obj, drs_obj)
        self.assertAlmostEqual(np_obj, a2dr_obj)
        for i in range(N):
            self.assertItemsAlmostEqual(np_beta[i], drs_beta[i])
            self.assertItemsAlmostEqual(np_beta[i], a2dr_beta[i])

    def test_infeas(self):
        # a modified non-negative least squares example with infeasible linear constraints
        m, n = 150, 300
        density = 0.001
        X = sparse.random(m, n, density=density, data_rvs=np.random.randn)
        y = np.random.randn(m)

        # Convert problem to standard form.
        prox_list = [lambda v, t: prox_sum_squares_affine(v, t, F=X, g=y), 
                     prox_nonneg_constr]
        Z = sparse.eye(n);
        e1 = sparse.lil_matrix((1,n))
        e1[0,0] = 1 
        A1 = sparse.bmat([[Z], [e1]])
        A2 = sparse.bmat([[-Z], [-e1]])
        A_list = [A1, A2]
        b = np.zeros(n+1)
        b[0] = 1
        b[-1] = -1

        # Solve with DRS.
        drs_result = a2dr(prox_list, A_list, b, anderson=False)
        # Solve with A2DR.
        a2dr_result = a2dr(prox_list, A_list, b, anderson=True)
        self.assertTrue(drs_result == 
            {"x_vals": None, "primal": None, "dual": None, "num_iters": None, "solve_time": None} 
            and a2dr_result == 
            {"x_vals": None, "primal": None, "dual": None, "num_iters": None, "solve_time": None})



import numpy as np
from scipy import sparse

def prox_scale(prox, *args, **kwargs):
    """Given the proximal operator of a function :math:`f`, returns the proximal operator of :math:`g` defined as
    .. math::
        g(x) = `tf(ax-b) + <c,x> + d\\|x\\|_F^2`,
    where :math:`t > 0`, :math:`a \\neq 0` is a scaling term, :math:`b` is an offset, :math:`c` is a linear multiplier,
    and :math:`d \\geq 0` is a quadratic multiplier.
    :param prox: Function handle of a proximal operator that takes as input a vector/matrix :math:`v` and a scalar
    :math:`t > 0`, and outputs
    .. math::
        `prox_{tf}(v) = \\min_x f(x) + \\frac{1}{2t}\\|x - v\\|_F^2}`,
    the proximal operator of :math:`tf` evaluated at :math:`v`, where :math:`f` is an arbitrary function.
    :param scale: Scaling term :math:`a \\neq 0`. Defaults to 1.
    :param offset: Offset term :math:`b`. Defaults to 0.
    :param lin_term: Linear term :math:`c`. If a scalar is given, then :math:`c` is its vectorization. Defaults to 0.
    :param quad_term: Quadratic term :math:`d \\geq 0`. Defaults to 0.
    :return: Function handle for the proximal operator of :math:`g`.
    """
    scale = kwargs.pop("scale", 1.0)
    offset = kwargs.pop("offset", 0)
    lin_term = kwargs.pop("lin_term", 0)
    quad_term = kwargs.pop("quad_term", 0)

    if not np.isscalar(scale) or scale == 0:
        raise ValueError("scale must be a non-zero scalar.")
    if not np.isscalar(quad_term) or quad_term < 0:
        raise ValueError("quad_term must be a non-negative scalar.")

    def prox_new(v, t):
        # if sparse.issparse(v):
        #     if not ((lin_term == 0 or sparse.issparse(lin_term)) and \
        #              (quad_term == 0 or sparse.issparse(quad_term))):
        #         v = v.todense()
        v_new = scale*(v - lin_term)/(2*quad_term + 1) - offset
        t_new = t*scale**2/(2*quad_term + 1)
        return (prox(v_new, t_new, *args, **kwargs) + offset)/scale
    return prox_new



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from cvxpy import *

from a2dr.proximal.projection import *
from a2dr.tests.base_test import BaseTest

class TestProjection(BaseTest):
    """Unit tests for projections"""

    def setUp(self):
        np.random.seed(1)
        self.TOLERANCE = 1e-6
        self.c = np.random.randn()
        self.v = np.random.randn(100)
        self.B = np.random.randn(50,10)

        self.radii = [0, 1, 2.5]
        self.radii += [np.abs(self.c) + self.TOLERANCE]
        self.radii += [np.max(np.abs(self.v)) + self.TOLERANCE,
                       (np.max(np.abs(self.v)) + np.min(np.abs(self.v)))/2]
        self.radii += [np.max(np.abs(self.B)) + self.TOLERANCE,
                       (np.max(np.abs(self.B)) + np.min(np.abs(self.B)))/2]

    def check_simplex(self, val, radius=1, method="bisection", places=4, *args, **kwargs):
        # Solve with CVXPY.
        x_var = Variable() if np.isscalar(val) else Variable(val.shape)
        obj = Minimize(sum_squares(x_var - val))
        constr = [x_var >= 0, sum(x_var) == radius]
        Problem(obj, constr).solve(*args, **kwargs)

        # Solve with projection algorithm.
        x_proj = proj_simplex(val, r=radius, method=method)
        self.assertItemsAlmostEqual(x_var.value, x_proj, places=places)

    def check_l1_ball(self, val, radius=1, method="bisection", places=4, *args, **kwargs):
        # Solve with CVXPY.
        x_var = Variable() if np.isscalar(val) else Variable(val.shape)
        obj = Minimize(sum_squares(x_var - val))
        constr = [norm1(x_var) <= radius]
        Problem(obj, constr).solve(*args, **kwargs)

        # Solve with projection algorithm.
        x_proj = proj_l1(val, r=radius, method=method)
        self.assertItemsAlmostEqual(x_var.value, x_proj, places=places)

    def test_simplex(self):
        for radius in self.radii:
            for method in ["bisection", "sorted"]:
                self.check_simplex(self.c, radius, method)
                self.check_simplex(self.v, radius, method)
                self.check_simplex(self.B, radius, method)

    def test_l1_ball(self):
        for radius in self.radii:
            for method in ["bisection", "sorted"]:
                self.check_l1_ball(self.c, radius, method, solver='ECOS')
                self.check_l1_ball(self.v, radius, method)
                self.check_l1_ball(self.B, radius, method)



import numpy as np
from scipy.optimize import bisect

TOLERANCE = 1e-6

def proj_l1(x, r = 1, method = "bisection"):
    if np.isscalar(x):
        x = np.array([x])
    if not np.isscalar(r) or r < 0:
        raise ValueError("r must be a non-negative scalar.")
    if np.linalg.norm(x,1) <= r:
        return x
    else:
        beta = proj_simplex(np.abs(x), r, method)
        return np.sign(x) * beta

def proj_simplex(x, r = 1, method = "bisection"):
    """Project x onto a simplex with upper bound r.
       Bisection: Liu and Ye (2009). "Efficient Euclidean Projections in Linear Time." Sect. 2.1.
            https://icml.cc/Conferences/2009/papers/123.pdf
       Efficient: Duchi et al (2008). "Efficient Projections onto the l1-Ball for Learning in High Dimensions."
            Fig. 1 and Sect. 4.
            https://stanford.edu/~jduchi/projects/DuchiShSiCh08.pdf
    """
    if np.isscalar(x):
        x = np.array([x])
    if not np.isscalar(r) or r < 0:
        raise ValueError("r must be a non-negative scalar.")
    elif r == 0:
        return np.zeros(x.shape)

    if method == "bisection":
        c_min = np.min(x) - (r + TOLERANCE)/x.size
        c_max = np.max(x) + (r + TOLERANCE)/x.size
        c_star = bisect(lambda c: np.sum(np.maximum(x - c, 0)) - r, c_min, c_max)
    elif method == "sorted":
        x_decr = np.sort(x, axis = None)[::-1]
        x_cumsum = np.cumsum(x_decr)
        denom = np.arange(1, x_decr.size + 1)
        theta = (x_cumsum - r)/denom
        x_diff = x_decr - theta
        idx = np.max(np.argwhere(x_diff > 0).ravel())
        c_star = theta[idx]
    else:
        raise ValueError("method must be either 'bisection' or 'sorted'.")
    return np.maximum(x - c_star, 0)



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy import sparse
from cvxpy import *
from a2dr.proximal import *
from a2dr.tests.base_test import BaseTest

class TestProximal(BaseTest):
    """Unit tests for proximal operators"""

    def setUp(self):
        np.random.seed(1)
        self.TOLERANCE = 1e-6
        self.SCS_TOLERANCE = 1e-6 #1e-8
        self.SCS_MAXITER = 10000
        self.t = 5*np.abs(np.random.randn()) + self.TOLERANCE
        self.c = np.random.randn()
        self.v = np.random.randn(100)
        self.v_small = np.random.randn(10)

        self.B = np.random.randn(50,10)
        self.B_small = np.random.randn(10,5)
        self.B_square = np.random.randn(10,10)

        self.B_symm = np.random.randn(10,10)
        self.B_symm = (self.B_symm + self.B_symm.T) / 2.0
        self.B_psd = np.random.randn(10,10)
        self.B_psd = self.B_psd.T.dot(self.B_psd)

        self.u_sparse = sparse.random(100,1)
        self.u_dense = self.u_sparse.todense()
        self.C_sparse = sparse.random(50,10)
        self.C_dense = self.C_sparse.todense()
        self.C_square_sparse = sparse.random(50,50)
        self.C_square_dense = self.C_square_sparse.todense()

    def prox_cvxpy(self, v, fun, constr_fun = None, t = 1, scale = 1, offset = 0, lin_term = 0, quad_term = 0, *args, **kwargs):
        x_var = Variable() if np.isscalar(v) else Variable(v.shape)
        expr = t * fun(scale * x_var - offset) + sum(multiply(lin_term, x_var)) + quad_term * sum_squares(x_var)
        constrs = [] if constr_fun is None else constr_fun(scale * x_var - offset)
        prob = Problem(Minimize(expr + 0.5 * sum_squares(x_var - v)), constrs)
        prob.solve(*args, **kwargs)
        return x_var.value

    def check_composition(self, prox, fun, v_init, places = 3, symm=False, *args, **kwargs):
        x_a2dr = prox(v_init)
        x_cvxpy = self.prox_cvxpy(v_init, fun, *args, **kwargs)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = places)

        x_a2dr = prox(v_init, t = self.t)
        x_cvxpy = self.prox_cvxpy(v_init, fun, t = self.t, *args, **kwargs)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = places)

        x_a2dr = prox(v_init, scale = -1)
        x_cvxpy = self.prox_cvxpy(v_init, fun, scale = -1, *args, **kwargs)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = places)

        x_a2dr = prox(v_init, scale = 2, offset = 0.5)
        x_cvxpy = self.prox_cvxpy(v_init, fun, scale = 2, offset = 0.5, *args, **kwargs)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = places)

        x_a2dr = prox(v_init, t = self.t, scale = 2, offset = 0.5, lin_term = 1.5, quad_term = 2.5)
        x_cvxpy = self.prox_cvxpy(v_init, fun, t = self.t, scale = 2, offset = 0.5, lin_term = 1.5, quad_term = 2.5,
                                  *args, **kwargs)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = places)

        if np.isscalar(v_init):
            offset = np.random.randn()
            lin_term = np.random.randn()
        else:
            offset = np.random.randn(*v_init.shape)
            lin_term = np.random.randn(*v_init.shape)
            if symm:
            # symmetrization: useful for -logdet, etc.
                offset = (offset + offset.T)/2
                lin_term = (lin_term + lin_term.T)/2
        x_a2dr = prox(v_init, t = self.t, scale = 0.5, offset = offset, lin_term = lin_term, quad_term = 2.5)
        x_cvxpy = self.prox_cvxpy(v_init, fun, t = self.t, scale = 0.5, offset = offset, lin_term = lin_term,
                                  quad_term = 2.5, *args, **kwargs)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = places)

    def check_elementwise(self, prox, places = 4):
        # Vector input.
        x_vec1 = prox(self.v_small)
        x_vec2 = np.array([prox(self.v_small[i]) for i in range(self.v_small.shape[0])])
        self.assertItemsAlmostEqual(x_vec1, x_vec2, places = places)

        x_vec1 = prox(self.v_small, t = self.t)
        x_vec2 = np.array([prox(self.v_small[i], t = self.t) for i in range(self.v_small.shape[0])])
        self.assertItemsAlmostEqual(x_vec1, x_vec2, places = places)

        offset = np.random.randn(*self.v_small.shape)
        lin_term = np.random.randn(*self.v_small.shape)
        x_vec1 = prox(self.v_small, t = self.t, scale = 0.5, offset = offset, lin_term = lin_term, quad_term = 2.5)
        x_vec2 = np.array([prox(self.v_small[i], t = self.t, scale = 0.5, offset = offset[i], lin_term = lin_term[i], \
                                quad_term = 2.5) for i in range(self.v_small.shape[0])])
        self.assertItemsAlmostEqual(x_vec1, x_vec2, places = places)

        # Matrix input.
        x_mat1 = prox(self.B_small)
        x_mat2 = [[prox(self.B_small[i,j]) for j in range(self.B_small.shape[1])] for i in range(self.B_small.shape[0])]
        x_mat2 = np.array(x_mat2)
        self.assertItemsAlmostEqual(x_mat1, x_mat2, places = places)

        x_mat1 = prox(self.B_small, t = self.t)
        x_mat2 = [[prox(self.B_small[i,j], t = self.t) for j in range(self.B_small.shape[1])] \
                    for i in range(self.B_small.shape[0])]
        x_mat2 = np.array(x_mat2)
        self.assertItemsAlmostEqual(x_mat1, x_mat2, places = places)

        offset = np.random.randn(*self.B_small.shape)
        lin_term = np.random.randn(*self.B_small.shape)
        x_mat1 = prox(self.B_small, t = self.t, scale = 0.5, offset = offset, lin_term = lin_term, quad_term = 2.5)
        x_mat2 = [[prox(self.B_small[i,j], t = self.t, scale = 0.5, offset = offset[i,j], lin_term = lin_term[i,j], \
                        quad_term = 2.5) for j in range(self.B_small.shape[1])] for i in range(self.B_small.shape[0])]
        x_mat2 = np.array(x_mat2)
        self.assertItemsAlmostEqual(x_mat1, x_mat2, places = places)

    def check_sparsity(self, prox, places = 4, check_vector = True, check_matrix = True, matrix_type = "general"):
        if check_vector:
            # Vector input.
            x_vec1 = prox(self.u_sparse)
            x_vec2 = prox(self.u_dense)
            self.assertTrue(sparse.issparse(x_vec1))
            self.assertItemsAlmostEqual(x_vec1.todense(), x_vec2, places = places)

            x_vec1 = prox(self.u_sparse, t=self.t)
            x_vec2 = prox(self.u_dense, t=self.t)
            self.assertTrue(sparse.issparse(x_vec1))
            self.assertItemsAlmostEqual(x_vec1.todense(), x_vec2, places = places)

            offset = sparse.random(*self.u_sparse.shape)
            lin_term = sparse.random(*self.u_sparse.shape)
            x_vec1 = prox(self.u_sparse, t=self.t, scale=0.5, offset=offset, lin_term=lin_term, quad_term=2.5)
            x_vec2 = prox(self.u_dense, t=self.t, scale=0.5, offset=offset, lin_term=lin_term, quad_term=2.5)
            self.assertTrue(sparse.issparse(x_vec1))
            self.assertItemsAlmostEqual(x_vec1.todense(), x_vec2, places = places)

        if check_matrix:
            if matrix_type == "general":
                C_sparse = self.C_sparse
                C_dense = self.C_dense
            elif matrix_type == "square":
                C_sparse = self.C_square_sparse
                C_dense = self.C_square_dense
            else:
                raise ValueError("matrix_type must be 'general' or 'square'")

            # Matrix input.
            x_mat1 = prox(C_sparse)
            x_mat2 = prox(C_dense)
            self.assertTrue(sparse.issparse(x_mat1))
            self.assertItemsAlmostEqual(x_mat1.todense(), x_mat2, places = places)

            x_mat1 = prox(C_sparse, t=self.t)
            x_mat2 = prox(C_dense, t=self.t)
            self.assertTrue(sparse.issparse(x_mat1))
            self.assertItemsAlmostEqual(x_mat1.todense(), x_mat2, places = places)

            offset = sparse.random(*C_sparse.shape)
            lin_term = sparse.random(*C_sparse.shape)
            x_mat1 = prox(C_sparse, t=self.t, scale=0.5, offset=offset, lin_term=lin_term, quad_term=2.5)
            x_mat2 = prox(C_dense, t=self.t, scale=0.5, offset=offset, lin_term=lin_term, quad_term=2.5)
            self.assertTrue(sparse.issparse(x_mat1))
            self.assertItemsAlmostEqual(x_mat1.todense(), x_mat2, places = places)

    def test_box_constr(self):
        # Projection onto a random interval.
        lo = np.random.randn()
        hi = lo + 5*np.abs(np.random.randn())

        x_a2dr = prox_box_constr(self.v, self.t, v_lo = lo, v_hi = hi)
        self.assertTrue(np.all(lo - self.TOLERANCE <= x_a2dr) and np.all(x_a2dr <= hi + self.TOLERANCE))

        # Projection onto a random interval with affine composition.
        scale = 2 * np.abs(np.random.randn()) + self.TOLERANCE
        if np.random.rand() < 0.5:
            scale = -scale
        offset = np.random.randn(*self.v.shape)
        lin_term = np.random.randn(*self.v.shape)
        quad_term = np.abs(np.random.randn())
        x_a2dr = prox_box_constr(self.v, self.t, v_lo = lo, v_hi = hi, scale = scale, offset = offset, \
                                    lin_term = lin_term, quad_term = quad_term)
        x_scaled = scale*x_a2dr - offset
        self.assertTrue(np.all(lo - self.TOLERANCE <= x_scaled) and np.all(x_scaled <= hi + self.TOLERANCE))

        # Common box intervals.
        bounds = [(0, 0), (-1, 1), (0, np.inf), (-np.inf, 0)]
        for bound in bounds:
            lo, hi = bound
            # Elementwise consistency tests.
            self.check_elementwise(lambda v, *args, **kwargs: prox_box_constr(v, v_lo = lo, v_hi = hi, *args, **kwargs))

            # Sparsity consistency tests.
            self.check_sparsity(lambda v, *args, **kwargs: prox_box_constr(v, v_lo = lo, v_hi = hi, *args, **kwargs))

            # General composition tests.
            self.check_composition(lambda v, *args, **kwargs: prox_box_constr(v, v_lo = lo, v_hi = hi, *args, **kwargs),
                                   lambda x: 0, self.v, constr_fun = lambda x: [lo <= x, x <= hi])
            self.check_composition(lambda v, *args, **kwargs: prox_box_constr(v, v_lo = lo, v_hi = hi, *args, **kwargs),
                                   lambda x: 0, self.B, constr_fun = lambda x: [lo <= x, x <= hi])

        # Optimal control term: f(x) = I(||x||_{\infty} <= 1) = I(-1 <= x <= 1).
        x_a2dr = prox_box_constr(self.v, self.t, v_lo = -1, v_hi = 1)
        x_cvxpy = self.prox_cvxpy(self.v, lambda x: 0, constr_fun = lambda x: [norm_inf(x) <= 1], t = self.t)
        self.assertTrue(np.all(-1 - self.TOLERANCE <= x_a2dr) and np.all(x_a2dr <= 1 + self.TOLERANCE))
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = 3)

    def test_nonneg_constr(self):
        x_a2dr = prox_nonneg_constr(self.v, self.t, scale = -2, offset = 0)
        self.assertTrue(np.all(-2*x_a2dr >= -self.TOLERANCE))

        scale = 2*np.abs(np.random.randn()) + self.TOLERANCE
        if np.random.rand() < 0.5:
            scale = -scale
        offset = np.random.randn(*self.v.shape)
        lin_term = np.random.randn(*self.v.shape)
        quad_term = np.abs(np.random.randn())

        x_a2dr = prox_nonneg_constr(self.v, self.t, scale = scale, offset = offset, lin_term = lin_term, \
                                    quad_term = quad_term)
        self.assertTrue(np.all(scale*x_a2dr - offset) >= -self.TOLERANCE)

        # Elementwise consistency tests.
        self.check_elementwise(prox_nonneg_constr)

        # Sparsity consistency tests.
        self.check_sparsity(prox_nonneg_constr)

        # General composition tests.
        self.check_composition(prox_nonneg_constr, lambda x: 0, self.v, constr_fun = lambda x: [x >= 0])
        self.check_composition(prox_nonneg_constr, lambda x: 0, self.B, constr_fun = lambda x: [x >= 0])

        # Non-negative least squares term: f(x) = I(x >= 0).
        x_a2dr = prox_nonneg_constr(self.v, self.t)
        x_cvxpy = self.prox_cvxpy(self.v, lambda x: 0, constr_fun = lambda x: [x >= 0], t = self.t)
        self.assertTrue(np.all(x_a2dr >= -self.TOLERANCE))
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = 3)

    def test_nonpos_constr(self):
        x_a2dr = prox_nonpos_constr(self.v, self.t)
        self.assertTrue(np.all(x_a2dr <= self.TOLERANCE))

        x_a2dr = prox_nonpos_constr(self.v, self.t, scale=-2, offset=0)
        self.assertTrue(np.all(-2*x_a2dr <= self.TOLERANCE))

        scale = 2 * np.abs(np.random.randn()) + self.TOLERANCE
        if np.random.rand() < 0.5:
            scale = -scale
        offset = np.random.randn(*self.v.shape)
        lin_term = np.random.randn(*self.v.shape)
        quad_term = np.abs(np.random.randn())

        x_a2dr = prox_nonpos_constr(self.v, self.t, scale=scale, offset=offset, lin_term=lin_term, \
                                    quad_term=quad_term)
        self.assertTrue(np.all(scale * x_a2dr - offset) <= self.TOLERANCE)

        # Elementwise consistency tests.
        self.check_elementwise(prox_nonpos_constr)

        # Sparsity consistency tests.
        self.check_sparsity(prox_nonpos_constr)

        # General composition tests.
        self.check_composition(prox_nonpos_constr, lambda x: 0, self.v, constr_fun=lambda x: [x <= 0])
        self.check_composition(prox_nonpos_constr, lambda x: 0, self.B, constr_fun=lambda x: [x <= 0])

    def test_psd_cone(self):
        # Projection onto the PSD cone.
        B_a2dr = prox_psd_cone(self.B_symm, self.t)
        self.assertTrue(np.all(np.linalg.eigvals(B_a2dr) >= -self.TOLERANCE))

        # Projection onto the PSD cone with affine composition.
        scale = 2 * np.abs(np.random.randn()) + self.TOLERANCE
        if np.random.rand() < 0.5:
            scale = -scale
        offset = np.random.randn(*self.B_symm.shape)
        lin_term = np.random.randn(*self.B_symm.shape)
        quad_term = np.abs(np.random.randn())
        B_a2dr = prox_psd_cone(self.B_symm, self.t, scale = scale, offset = offset, lin_term = lin_term, \
                                  quad_term = quad_term)
        B_scaled = scale*B_a2dr - offset
        self.assertTrue(np.all(np.linalg.eigvals(B_scaled) >= -self.TOLERANCE))

        # Simple composition.
        B_a2dr = prox_psd_cone(self.B_symm, t = self.t, scale = 2, offset = 0.5, lin_term = 1.5, quad_term = 2.5)
        B_cvxpy = self.prox_cvxpy(self.B_symm, lambda X: 0, constr_fun = lambda X: [X >> 0], t = self.t, scale = 2, \
                                  offset = 0.5, lin_term = 1.5, quad_term = 2.5)
        self.assertItemsAlmostEqual(B_a2dr, B_cvxpy)

    def test_soc(self):
        # Projection onto the SOC.
        x_a2dr = prox_soc(self.v, self.t)
        self.assertTrue(np.linalg.norm(x_a2dr[:-1],2) <= x_a2dr[-1] + self.TOLERANCE)

        # Projection onto the SOC with affine composition.
        x_a2dr = prox_soc(self.v, self.t, scale=2, offset=0.5)
        x_scaled = 2*x_a2dr - 0.5
        self.assertTrue(np.linalg.norm(x_scaled[:-1],2) <= x_scaled[-1] + self.TOLERANCE)

        scale = 2 * np.abs(np.random.randn()) + self.TOLERANCE
        if np.random.rand() < 0.5:
            scale = -scale
        offset = np.random.randn(*self.v.shape)
        lin_term = np.random.randn(*self.v.shape)
        quad_term = np.abs(np.random.randn())

        x_a2dr = prox_soc(self.v, self.t, scale=scale, offset=offset, lin_term=lin_term, quad_term=quad_term)
        x_scaled = scale*x_a2dr - offset
        self.assertTrue(np.linalg.norm(x_scaled[:-1], 2) <= x_scaled[-1] + self.TOLERANCE)

        # Sparsity consistency tests.
        self.check_sparsity(prox_soc, check_matrix = False)

        # General composition tests.
        self.check_composition(prox_soc, lambda x: 0, self.v, constr_fun = lambda x: [SOC(x[-1], x[:-1])], \
                               solver = "SCS", eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

    def test_abs(self):
        # Elementwise consistency tests.
        self.check_elementwise(prox_abs)

        # General composition tests.
        self.check_composition(prox_abs, cvxpy.abs, self.c)
        self.check_composition(prox_abs, lambda x: sum(abs(x)), self.v)
        self.check_composition(prox_abs, lambda x: sum(abs(x)), self.B)

    def test_constant(self):
        # Elementwise consistency tests.
        self.check_elementwise(prox_constant)

        # Sparsity consistency tests.
        self.check_sparsity(prox_constant)

        # General composition tests.
        self.check_composition(prox_constant, lambda x: 0, self.c)
        self.check_composition(prox_constant, lambda x: 0, self.v)
        self.check_composition(prox_constant, lambda x: 0, self.B)

    def test_exp(self):
        # Elementwise consistency tests.
        self.check_elementwise(prox_exp)

        # General composition tests.
        self.check_composition(prox_exp, cvxpy.exp, self.c, solver = "SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_exp, lambda x: sum(exp(x)), self.v, solver = "SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_exp, lambda x: sum(exp(x)), self.B, places=2, solver = "SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

    def test_huber(self):
        for M in [0, 0.5, 1, 2]:
            # Elementwise consistency tests.
            self.check_elementwise(lambda v, *args, **kwargs: prox_huber(v, *args, **kwargs, M = M))

            # Sparsity consistency tests.
            self.check_sparsity(lambda v, *args, **kwargs: prox_huber(v, *args, **kwargs, M = M))

            # Scalar input.
            self.check_composition(lambda v, *args, **kwargs: prox_huber(v, M = M, *args, **kwargs),
                                   lambda x: huber(x, M = M), self.c)
            # Vector input.
            self.check_composition(lambda v, *args, **kwargs: prox_huber(v, M = M, *args, **kwargs),
                                   lambda x: sum(huber(x, M = M)), self.v)
            # Matrix input.
            self.check_composition(lambda v, *args, **kwargs: prox_huber(v, M = M, *args, **kwargs),
                                   lambda x: sum(huber(x, M = M)), self.B)

    def test_identity(self):
        # Elementwise consistency tests.
        self.check_elementwise(prox_identity)

        # General composition tests.
        self.check_composition(prox_identity, lambda x: x, self.c)
        self.check_composition(prox_identity, lambda x: sum(x), self.v)
        self.check_composition(prox_identity, lambda x: sum(x), self.B)

    def test_logistic(self):
        # General composition tests.
        # self.check_composition(prox_logistic, lambda x: logistic(x), self.c, solver='ECOS')
        self.check_composition(prox_logistic, lambda x: logistic(x), self.c, solver = 'SCS', 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_logistic, lambda x: sum(logistic(x)), self.v, solver = 'SCS', 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_logistic, lambda x: sum(logistic(x)), self.B, solver = "SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

        # Simple logistic function: f(x) = \sum_i log(1 + exp(-y_i*x_i)).
        y = np.random.randn(*self.v.shape)
        self.check_composition(lambda v, *args, **kwargs: prox_logistic(v, y = y, *args, **kwargs),
                               lambda x: sum(logistic(-multiply(y,x))), self.v, places = 2, solver = "SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

        # Multi-task logistic regression term: f(B) = \sum_i log(1 + exp(-Y_{ij}*B_{ij}).
        Y_mat = np.random.randn(*self.B.shape)
        B_a2dr = prox_logistic(self.B, t = self.t, y = Y_mat)
        B_cvxpy = self.prox_cvxpy(self.B, lambda B: sum(logistic(-multiply(Y_mat,B))), t = self.t, solver = "SCS", 
                                  eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.assertItemsAlmostEqual(B_a2dr, B_cvxpy, places = 2)

        self.check_composition(lambda v, *args, **kwargs: prox_logistic(v, y = Y_mat, *args, **kwargs),
                               lambda B: sum(logistic(-multiply(Y_mat,B))), self.B, places = 2, solver = "SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

    def test_pos(self):
        # Elementwise consistency tests.
        self.check_elementwise(prox_pos)

        # Sparsity consistency tests.
        self.check_sparsity(prox_pos)

        # General composition tests.
        self.check_composition(prox_pos, cvxpy.pos, self.c)
        self.check_composition(prox_pos, lambda x: sum(pos(x)), self.v)
        self.check_composition(prox_pos, lambda x: sum(pos(x)), self.B)

    def test_neg(self):
        # Elementwise consistency tests.
        self.check_elementwise(prox_neg)

        # Sparsity consistency tests.
        self.check_sparsity(prox_neg)

        # General composition tests.
        self.check_composition(prox_neg, cvxpy.neg, self.c)
        self.check_composition(prox_neg, lambda x: sum(neg(x)), self.v)
        self.check_composition(prox_neg, lambda x: sum(neg(x)), self.B)

    def test_neg_entr(self):
        # Elementwise consistency tests.
        self.check_elementwise(prox_neg_entr)

        # General composition tests.
        self.check_composition(prox_neg_entr, lambda x: -entr(x), self.c, solver="SCS",
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_neg_entr, lambda x: sum(-entr(x)), self.v, solver="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_neg_entr, lambda x: sum(-entr(x)), self.B, places=2, solver="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

    def test_neg_log(self):
        # Elementwise consistency tests.
        self.check_elementwise(prox_neg_log)

        # General composition tests.
        self.check_composition(prox_neg_log, lambda x: -log(x), self.c, solver='SCS', 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_neg_log, lambda x: sum(-log(x)), self.v, solver='SCS', 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_neg_log, lambda x: sum(-log(x)), self.B, places=2, solver="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

    def test_neg_log_det(self):
        # General composition tests.
        self.check_composition(prox_neg_log_det, lambda X: -log_det(X), self.B_symm, places=2, symm=True, solver="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_neg_log_det, lambda X: -log_det(X), self.B_psd, places=2, symm=True, solver="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

        # Sparse inverse covariance estimation term: f(B) = -log(det(B)) for symmetric positive definite B.
        B_spd = self.B_psd + np.eye(self.B_psd.shape[0])
        B_a2dr = prox_neg_log_det(B_spd, self.t)
        B_cvxpy = self.prox_cvxpy(B_spd, lambda X: -log_det(X), t=self.t, solver="SCS", 
                                  eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.assertItemsAlmostEqual(B_a2dr, B_cvxpy, places=2)

        # Sparse inverse covariance estimation term: f(B) = -log(det(B)) + tr(BQ) for symmetric positive definite B
        # and given matrix Q.
        Q = np.random.randn(*B_spd.shape)
        Q = (Q + Q.T)/2 # keep Q symmetric to be valid input for -logdet proximal
        B_a2dr = prox_neg_log_det(B_spd, self.t, lin_term = self.t*Q.T)   # tr(A^TB) = \sum_{ij} A_{ij}B_{ij}
        B_cvxpy = self.prox_cvxpy(B_spd, lambda X: -log_det(X) + trace(Q*X), t=self.t, solver="SCS", 
                                  eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.assertItemsAlmostEqual(B_a2dr, B_cvxpy, places=2)

    def test_max(self):
        # General composition tests.
        self.check_composition(prox_max, cvxpy.max, self.c)
        self.check_composition(prox_max, cvxpy.max, self.v)
        # self.check_composition(prox_max, cvxpy.max, self.B, solver = "SCS", 
        #                        eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_max, cvxpy.max, self.B)

    def test_kl(self):
        # General composition tests.
        u_c = np.random.rand() + 1e-8
        u_v = np.random.rand(*self.v.shape) + 1e-8
        u_B = np.random.rand(*self.B.shape) + 1e-8
        self.check_composition(lambda v, *args, **kwargs: prox_kl(v, u = u_c, *args, **kwargs), 
                               lambda x: sum(-entr(x)-x*log(u_c)), self.c, solver='SCS',
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(lambda v, *args, **kwargs: prox_kl(v, u = u_v, *args, **kwargs),
                               lambda x: sum(-entr(x)-multiply(x, log(u_v))), self.v, solver='SCS',
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(lambda v, *args, **kwargs: prox_kl(v, u = u_B, *args, **kwargs),  
                               lambda x: sum(-entr(x)-multiply(x, log(u_B))), self.B, solver='SCS',
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

    def test_norm1(self):
        # Sparsity consistency tests.
        self.check_sparsity(prox_norm1)

        # General composition tests.
        self.check_composition(prox_norm1, norm1, self.c)
        self.check_composition(prox_norm1, norm1, self.v)
        self.check_composition(prox_norm1, norm1, self.B)

        # l1 trend filtering term: f(x) = \alpha*||x||_1.
        alpha = 0.5 + np.abs(np.random.randn())
        x_a2dr = prox_norm1(self.v, t = alpha*self.t)
        x_cvxpy = self.prox_cvxpy(self.v, norm1, t = alpha*self.t)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = 4)

        # Sparse inverse covariance estimation term: f(B) = \alpha*||B||_1.
        B_symm_a2dr = prox_norm1(self.B_symm, t = alpha*self.t)
        B_symm_cvxpy = self.prox_cvxpy(self.B_symm, norm1, t = alpha*self.t)
        self.assertItemsAlmostEqual(B_symm_a2dr, B_symm_cvxpy, places = 4)

    def test_norm2(self):
        # Sparsity consistency tests.
        self.check_sparsity(prox_norm2)

        # General composition tests.
        self.check_composition(prox_norm2, norm2, self.c, solver ="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_norm2, norm2, self.v, solver ="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.check_composition(prox_norm2, lambda B: cvxpy.norm(B, 'fro'), self.B, solver ="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

        # f(x) = \alpha*||x||_2
        alpha = 0.5 + np.abs(np.random.randn())
        x_a2dr = prox_norm2(self.v, t = alpha*self.t)
        x_cvxpy = self.prox_cvxpy(self.v, norm2, t = alpha*self.t)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = 4)

    def test_norm_inf(self):
        # General composition tests.
        self.check_composition(prox_norm_inf, norm_inf, self.c)
        self.check_composition(prox_norm_inf, norm_inf, self.v)
        # self.check_composition(prox_norm_inf, norm_inf, self.B, solver='ECOS')
        self.check_composition(prox_norm_inf, norm_inf, self.B, solver ="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

        # f(x) = \alpha*||x||_{\infty}
        alpha = 0.5 + np.abs(np.random.randn())
        x_a2dr = prox_norm_inf(self.v, t = alpha*self.t)
        x_cvxpy = self.prox_cvxpy(self.v, norm_inf, t = alpha*self.t)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = 4)

    def test_norm_fro(self):
        # General composition tests.
        self.check_composition(prox_norm_fro, lambda X: cvxpy.norm(X,'fro'), self.B, solver='SCS', 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

    def test_norm_nuc(self):
        # General composition tests.
        self.check_composition(prox_norm_nuc, normNuc, self.B, solver='SCS', 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

        # Multi-task logistic regression term: f(B) = \beta*||B||_*.
        beta = 1.5 + np.abs(np.random.randn())
        B_a2dr = prox_norm_nuc(self.B, t = beta*self.t)
        B_cvxpy = self.prox_cvxpy(self.B, normNuc, t = beta*self.t, solver='SCS', 
                                  eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.assertItemsAlmostEqual(B_a2dr, B_cvxpy, places = 3)

    def test_group_lasso(self):
        # Sparsity consistency tests.
        self.check_sparsity(prox_group_lasso)

        # General composition tests.
        groupLasso = lambda B: sum([norm2(B[:,j]) for j in range(B.shape[1])])
        self.check_composition(prox_group_lasso, groupLasso, self.B, solver="SCS", 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

        # Multi-task logistic regression term: f(B) = \alpha*||B||_{2,1}.
        alpha = 1.5 + np.abs(np.random.randn())
        B_a2dr = prox_group_lasso(self.B, t = alpha*self.t)
        B_cvxpy = self.prox_cvxpy(self.B, groupLasso, t = alpha*self.t, solver="SCS", 
                                  eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)
        self.assertItemsAlmostEqual(B_a2dr, B_cvxpy, places = 3)

        # Compare with taking l2-norm separately on each column.
        B_norm2 = [prox_norm2(self.B[:,j], t = alpha*self.t) for j in range(self.B.shape[1])]
        B_norm2 = np.vstack(B_norm2)
        self.assertItemsAlmostEqual(B_a2dr, B_norm2, places = 3)

    def test_sigma_max(self):
        # General composition tests.
        self.check_composition(prox_sigma_max, sigma_max, self.B, solver='SCS', 
                               eps=self.SCS_TOLERANCE, max_iters=self.SCS_MAXITER)

    def test_sum_squares(self):
        # Sparsity consistency tests.
        self.check_sparsity(prox_sum_squares)

        # General composition tests.
        self.check_composition(prox_sum_squares, sum_squares, self.v)
        self.check_composition(prox_sum_squares, sum_squares, self.B)

        # Optimal control term: f(x) = ||x||_2^2.
        x_a2dr = prox_sum_squares(self.v, t = self.t)
        x_cvxpy = self.prox_cvxpy(self.v, sum_squares, t = self.t)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = 4)

        # l1 trend filtering term: f(x) = (1/2)*||x - y||_2^2 for given y.
        y = np.random.randn(*self.v.shape)
        x_a2dr = prox_sum_squares(self.v, t = 0.5*self.t, offset = y)
        x_cvxpy = self.prox_cvxpy(self.v, sum_squares, t = 0.5*self.t, offset = y)
        self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = 4)

    def test_sum_squares_affine(self):
        # Scalar terms.
        F = np.random.randn()
        g = np.random.randn()
        v = np.random.randn()

        self.check_composition(lambda v, *args, **kwargs: prox_sum_squares_affine(v, F = F, g = g, method ="lsqr",
                                                                                  *args, **kwargs), lambda x: sum_squares(F*x - g), v)
        self.check_composition(lambda v, *args, **kwargs: prox_sum_squares_affine(v, F = F, g = g, method ="lstsq",
                                                                                  *args, **kwargs), lambda x: sum_squares(F*x - g), v)

        # Simple sum of squares: f(x) = ||x||_2^2.
        n = 100
        F = np.eye(n)
        g = np.zeros(n)
        v = np.random.randn(n)
        F_sparse = sparse.eye(n)
        g_sparse = sparse.csr_matrix((n,1))

        for method in ["lsqr", "lstsq"]:
            self.check_composition(lambda v, *args, **kwargs: prox_sum_squares_affine(v, F = F, g = g, \
                                        method = method, *args, **kwargs), lambda x: sum_squares(F*x - g), v)
            self.check_composition(lambda v, *args, **kwargs: prox_sum_squares_affine(v, F = F_sparse, g = g_sparse, \
                                        method = method, *args, **kwargs), lambda x: sum_squares(F*x - g), v)

        # Non-negative least squares term: f(x) = ||Fx - g||_2^2.
        m = 1000
        n = 100
        F = 10 + 5*np.random.randn(m,n)
        x = 2*np.random.randn(n)
        g = F.dot(x) + 0.01*np.random.randn(m)
        v = np.random.randn(n)

        for method in ["lsqr", "lstsq"]:
            x_a2dr = prox_sum_squares_affine(self.v, t = self.t, F = F, g = g, method = method)
            x_cvxpy = self.prox_cvxpy(self.v, lambda x: sum_squares(F*x - g), t = self.t)
            self.assertItemsAlmostEqual(x_a2dr, x_cvxpy, places = 4)

        # General composition tests.
        self.check_composition(lambda v, *args, **kwargs: prox_sum_squares_affine(v, F = F, g = g, method ="lsqr",
                                                                                  *args, **kwargs), lambda x: sum_squares(F*x - g), v)
        self.check_composition(lambda v, *args, **kwargs: prox_sum_squares_affine(v, F = F, g = g, method ="lstsq",
                                                                                  *args, **kwargs), lambda x: sum_squares(F*x - g), v)

    def test_quad_form(self):
        # Simple quadratic.
        v = np.random.randn(1)
        Q = np.array([[5]])
        self.check_composition(lambda v, *args, **kwargs: prox_quad_form(v, Q = Q, method = "lsqr", *args, **kwargs),
                               lambda x: quad_form(x, P = Q), v)
        self.check_composition(lambda v, *args, **kwargs: prox_quad_form(v, Q = Q, method = "lstsq", *args, **kwargs),
                               lambda x: quad_form(x, P = Q), v)

        # General composition tests.
        n = 10
        v = np.random.randn(n)
        Q = np.random.randn(n,n)
        Q = Q.T.dot(Q) + 0.5*np.eye(n)
        self.check_composition(lambda v, *args, **kwargs: prox_quad_form(v, Q = Q, method = "lsqr", *args, **kwargs),
                               lambda x: quad_form(x, P = Q), v)
        self.check_composition(lambda v, *args, **kwargs: prox_quad_form(v, Q = Q, method = "lstsq", *args, **kwargs),
                               lambda x: quad_form(x, P = Q), v)

    def test_trace(self):
        # Sparsity consistency tests.
        C = sparse.random(*self.C_square_sparse.shape)
        self.check_sparsity(prox_trace, check_vector = False, matrix_type = "square")
        self.check_sparsity(lambda B, *args, **kwargs: prox_trace(B, C = C, *args, **kwargs), check_vector = False, \
                            matrix_type = "square")

        # General composition tests.
        C = np.random.randn(*self.B.shape)
        self.check_composition(prox_trace, cvxpy.trace, self.B_square)
        self.check_composition(lambda B, *args, **kwargs: prox_trace(B, C = C, *args, **kwargs),
                               lambda X: cvxpy.trace(C.T * X), self.B)

        # Sparse inverse covariance estimation term: f(B) = tr(BQ) for given symmetric positive semidefinite Q.
        Q = np.random.randn(*self.B_square.shape)
        Q = Q.T.dot(Q)
        B_a2dr = prox_trace(self.B_square, t = self.t, C = Q.T)   # tr(BQ) = tr(QB) = tr((Q^T)^TB).
        B_cvxpy = self.prox_cvxpy(self.B_square, lambda X: cvxpy.trace(X * Q), t = self.t)
        self.assertItemsAlmostEqual(B_a2dr, B_cvxpy, places = 4)



import numpy as np
from cvxpy import *
from scipy import sparse
from a2dr.proximal.composition import prox_scale

def prox_quad_form(v, t = 1, Q = None, method = "lsqr", *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = x^TQx` for symmetric
    :math:`Q \\succeq 0`, scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term,
    and d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0,
    and d = 0.
    """
    if np.isscalar(v):
        v = np.array([v])
    if Q is None:
        raise ValueError("Q must be a matrix.")
    elif np.isscalar(Q):
        Q = np.array([[Q]])

    if Q.shape[0] != Q.shape[1]:
        raise ValueError("Q must be a square matrix.")
    if Q.shape[0] != v.shape[0]:
        raise ValueError("Dimension mismatch: nrow(Q) != nrow(v).")
    if sparse.issparse(Q):
        Q_min_eigval = sparse.linalg.eigsh(Q, k=1, which="SA", return_eigenvectors=False)[0]
        if np.iscomplex(Q_min_eigval) or Q_min_eigval < 0:
            raise ValueError("Q must be a symmetric positive semidefinite matrix.")
    else:
        if not np.all(np.linalg.eigvalsh(Q) >= 0):
            raise ValueError("Q must be a symmetric positive semidefinite matrix.")
    if method not in ["lsqr", "lstsq"]:
        raise ValueError("method must be either 'lsqr' or 'lstsq'")
    return prox_scale(prox_quad_form_base, Q=Q, method=method, *args, **kwargs)(v, t)

def prox_sum_squares(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = \\sum_i x_i^2`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_sum_squares_base, *args, **kwargs)(v, t)

def prox_sum_squares_affine(v, t = 1, F = None, g = None, method = "lsqr", *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = \\|Fx - g\\|_2^2`
        for matrix F, vector g, scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term,
        and d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0,
        and d = 0.
        """
    if np.isscalar(v):
        v = np.array([v])
    if F is None:
        raise ValueError("F must be a matrix.")
    elif np.isscalar(F):
        F = np.array([[F]])
    if g is None:
        raise ValueError("g must be a vector.")
    elif np.isscalar(g):
        g = np.array([g])

    if F.shape[0] != g.shape[0]:
        raise ValueError("Dimension mismatch: nrow(F) != nrow(g).")
    if F.shape[1] != v.shape[0]:
        raise ValueError("Dimension mismatch: ncol(F) != nrow(v).")
    if method not in ["lsqr", "lstsq"]:
        raise ValueError("method must be either 'lsqr' or 'lstsq'.")
    return prox_scale(prox_sum_squares_affine_base, F=F, g=g, method=method, *args, **kwargs)(v, t)

def prox_qp(v, t = 1, Q = None, q = None, F = None, g = None, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = x^TQx+q^Tx+I{Fx<=g}`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if Q is None:
        raise ValueError("Q must be a matrix")
    if q is None:
        raise ValueError("q must be a vector")
    if F is None:
        raise ValueError("F must be a matrix")
    if g is None:
        raise ValueError("g must be a vector")
    if Q.shape[0] != Q.shape[1]:
        raise ValueError("Q must be square")
    if Q.shape[0] != q.shape[0]:
        raise ValueError("Dimension mismatch: nrow(Q) != nrow(q)")
    if Q.shape[1] != v.shape[0]:
        raise ValueError("Dimension mismatch: ncol(Q) != nrow(v)")
    if F.shape[0] != g.shape[0]:
        raise ValueError("Dimension mismatch: nrow(F) != nrow(g)")
    if F.shape[1] != v.shape[0]:
        raise ValueError("Dimension mismatch: ncol(F) != nrow(v)")
    return prox_scale(prox_qp_base(Q, q, F, g), *args, **kwargs)(v, t)

def prox_quad_form_base(v, t, Q, method = "lsqr"):
    """Proximal operator of :math:`f(x) = x^TQx`, where :math:`Q \\succeq 0` is a symmetric positive semidefinite matrix.
    """
    if method == "lsqr":
        # Q_min_eigval = spLA.eigsh(Q, k=1, which="SA", return_eigenvectors=False)[0]
        # if np.iscomplex(Q_min_eigval) or Q_min_eigval < 0:
        #    raise Exception("Q must be a symmetric positive semidefinite matrix.")
        Q = sparse.csr_matrix(Q)
        return sparse.linalg.lsqr(2*t*Q + sparse.eye(v.shape[0]), v, atol=1e-16, btol=1e-16)[0]
    elif method == "lstsq":
        # if not np.all(LA.eigvalsh(Q) >= 0):
        #    raise Exception("Q must be a symmetric positive semidefinite matrix.")
        return np.linalg.lstsq(2*t*Q + np.eye(v.shape[0]), v, rcond=None)[0]
    else:
        raise ValueError("method must be 'lsqr' or 'lstsq'")

def prox_sum_squares_base(v, t):
    """Proximal operator of :math:`f(x) = \\sum_i x_i^2`.
    """
    return v / (1.0 + 2*t)

def prox_sum_squares_affine_base(v, t, F, g, method = "lsqr"):
    """Proximal operator of :math:`f(x) = \\|Fx - g\\|_2^2`, where F is a matrix and g is a vector.
    """
    # if F.shape[0] != g.shape[0]:
    #    raise ValueError("Dimension mismatch: nrow(F) != nrow(g)")
    # if F.shape[1] != v.shape[0]:
    #    raise ValueError("Dimension mismatch: ncol(F) != nrow(v)")

    # Only works on dense vectors.
    if sparse.issparse(g):
        g = g.toarray()[:, 0]
    if sparse.issparse(v):
        v = v.toarray()[:, 0]

    n = v.shape[0]
    if method == "lsqr":
        F = sparse.csr_matrix(F)
        F_stack = sparse.vstack([F, 1/np.sqrt(2*t)*sparse.eye(n)])
        g_stack = np.concatenate([g, 1/np.sqrt(2*t)*v])
        return sparse.linalg.lsqr(F_stack, g_stack, atol=1e-16, btol=1e-16)[0]
    elif method == "lstsq":
        if sparse.issparse(F):
            F = F.todense()
        F_stack = np.vstack([F, 1/np.sqrt(2*t)*np.eye(n)])
        g_stack = np.concatenate([g, 1/np.sqrt(2*t)*v])
        return np.linalg.lstsq(F_stack, g_stack, rcond=None)[0]
    else:
        raise ValueError("method must be 'lsqr' or 'lstsq'")

def prox_qp_base(Q, q, F, g):
    # check warmstart/parameter mode -- make sure the problem reduction is only done once
    n = Q.shape[0]
    I = np.eye(n)
    v_par = Parameter(n)
    t_par = Parameter(nonneg=True)
    x = Variable(n)
    obj = quad_form(x, Q) + sum_squares(x)/2/t_par + (q-v_par/t_par)*x
    constr = [F * x <= g]
    prob = Problem(Minimize(obj), constr)
    def prox_qp1(v, t):
        v_par.value, t_par.value = v, t
        prob.solve()
        return x.value
    return prox_qp1



from setuptools import setup, find_packages
import codecs
import os.path

# code for single sourcing versions
# reference: https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

# read the contents of your README file
# reference: https://packaging.python.org/guides/making-a-pypi-friendly-readme/
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='a2dr',
      version=get_version("a2dr/__init__.py"),
      description='A Python package for type-II Anderson accelerated Douglas-Rachford splitting (A2DR).',
      url='https://github.com/cvxgrp/a2dr',
      author='Anqi Fu, Junzi Zhang, Stephen Boyd',
      author_email='contact.a2dr.solver@gmail.com',
      license='Apache License, Version 2.0',
      packages=find_packages(),
      install_requires=['matplotlib',
                        'cvxpy >= 1.0.25',
                        'numpy >= 1.16',
                        'scipy >= 1.2.1'],
      zip_safe=False,
      test_suite='nose.collector',
      tests_require=['nose'],
      long_description=long_description,
      long_description_content_type='text/markdown')



"""
Copyright 2019 Anqi Fu, Junzi Zhang

This file is part of A2DR.

A2DR is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

A2DR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with A2DR. If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import scipy as sp
from scipy import sparse

from a2dr import a2dr
from a2dr.proximal import prox_norm1, prox_sum_squares_affine
from a2dr.precondition import precondition
from a2dr.tests.base_test import BaseTest

class TestPrecondition(BaseTest):
    """Unit tests for preconditioning data before S-DRS"""

    def setUp(self):
        np.random.seed(1)
        self.MAX_ITERS = 1000

    def test_precond_l1_trend_filter(self):
        # Problem data.
        N = 2
        n0 = 2*10**4
        n = 2*n0-2
        m = n0-2
        y = np.random.randn(n)
        alpha = 0.1*np.linalg.norm(y, np.inf)

        # Form second difference matrix.
        D = sparse.lil_matrix(sparse.eye(n0))
        D.setdiag(-2, k = 1)
        D.setdiag(1, k = 2)
        D = D[:(n0-2),:]

        # Convert problem to standard form.
        # f_1(x_1) = (1/2)||y - x_1||_2^2, f_2(x_2) = \alpha*||x_2||_1.
        # A_1 = D, A_2 = -I_{n-2}, b = 0.
        prox_list = [lambda v, t: (t*y + v)/(t + 1.0), lambda v, t: prox_norm1(v, t = alpha*t)]
        A_list = [D, -sparse.eye(n0-2)]
        b = np.zeros(n0-2)

        b = np.random.randn(m)
        prox_list = [prox_norm1] * N
        A = sparse.csr_matrix(sparse.hstack(A_list))
        
        p_eq_list, A_eq_list, db, e = precondition(prox_list, A_list, b)
        A_eq = sparse.csr_matrix(sparse.hstack(A_eq_list))
        
        print(r'[Sanity Check]')
        print(r'\|A\|_2 = {}, \|DAE\|_2 = {}'.format(sparse.linalg.norm(A), sparse.linalg.norm(A_eq)))
        print(r'min(|A|) = {}, max(|A|) = {}, mean(|A|) = {}'.format(np.min(np.abs(A)), 
                                                                    np.max(np.abs(A)), sparse.csr_matrix.mean(np.abs(A))))
        print(r'min(|DAE|) = {}, max(|DAE|) = {}, mean(|DAE|) = {}'.format(np.min(np.abs(A_eq)), 
                                                                    np.max(np.abs(A_eq)), sparse.csr_matrix.mean(np.abs(A_eq))))

    def test_nnls(self):
        # Solve the non-negative least squares problem
        # Minimize (1/2)*||A*x - b||_2^2 subject to x >= 0.
        m = 100
        n = 10
        N = 1   # Number of nodes (split A row-wise)

        # Problem data.
        mu = 100
        sigma = 10
        X = mu + sigma*np.random.randn(m,n)
        y = mu + sigma*np.random.randn(m)

        # Solve with SciPy.
        sp_result = sp.optimize.nnls(X, y)
        sp_beta = sp_result[0]
        sp_obj = sp_result[1] ** 2  # SciPy objective is ||y - X\beta||_2.
        print("Scipy Objective:", sp_obj)
        print("SciPy Solution:", sp_beta)

        X_split = np.split(X, N)
        y_split = np.split(y, N)
        p_list = [lambda v, t: prox_sum_squares_affine(v, t, F=X_sub, g=y_sub, method="lsqr") \
                    for X_sub, y_sub in zip(X_split, y_split)]
        p_list += [lambda u, rho: np.maximum(u, 0)]   # Projection onto non-negative orthant.
        A_list = np.hsplit(np.eye(N*n), N) + [-np.vstack(N*(np.eye(n),))]
        b = np.zeros(N*n)

        # Solve with A2DR.
        a2dr_result = a2dr(p_list, A_list, b, anderson=True, precond=False, max_iter=self.MAX_ITERS)
        a2dr_beta = a2dr_result["x_vals"][-1]
        a2dr_obj = np.sum((y - X.dot(a2dr_beta))**2)
        print("A2DR Objective:", a2dr_obj)
        print("A2DR Solution:", a2dr_beta)
        self.assertAlmostEqual(sp_obj, a2dr_obj)
        self.assertItemsAlmostEqual(sp_beta, a2dr_beta, places=3)

        # Solve with preconditioned A2DR.
        cond_result = a2dr(p_list, A_list, b, anderson=True, precond=True, max_iter=self.MAX_ITERS)
        cond_beta = cond_result["x_vals"][-1]
        cond_obj = np.sum((y - X.dot(cond_beta))**2)
        print("Preconditioned A2DR Objective:", cond_obj)
        print("Preconditioned A2DR Solution:", cond_beta)
        self.assertAlmostEqual(sp_obj, cond_obj)
        self.assertItemsAlmostEqual(sp_beta, cond_beta, places=3)



import numpy as np
from scipy import sparse
from scipy.special import lambertw
from a2dr.proximal.interface import NUMPY_FUNS, SPARSE_FUNS, apply_to_nonzeros
from a2dr.proximal.composition import prox_scale

def prox_abs(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = |x|` applied elementwise
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_abs_base, *args, **kwargs)(v, t)

def prox_constant(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = c` for constant c,
    scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_constant_base, *args, **kwargs)(v, t)

def prox_exp(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = \\exp(x)` applied
    elementwise for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and
    d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and 
    d = 0.
    """
    return prox_scale(prox_exp_base, *args, **kwargs)(v, t)

def prox_huber(v, t = 1, M = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where
    .. math::
        f(x) =
            \\begin{cases}
                2M|x|-M^2 & \\text{for } |x| \\geq |M| \\\\
                      |x|^2 & \\text{for } |x| \\leq |M|
            \\end{cases}
    applied elementwise for scalar M > 0, t > 0, and the optional arguments are a = scale, b = offset, c = lin_term,
    and d = quad_term. We must have M > 0, t > 0, a = non-zero, and d >= 0. By default, M = 1, t = 1, a = 1, b = 0, 
    c = 0, and d = 0.
    """
    return prox_scale(prox_huber_base, M, *args, **kwargs)(v, t)

def prox_identity(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = x` applied elementwise
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_identity_base, *args, **kwargs)(v, t)

def prox_neg(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = -\\min(x,0)` applied
    elementwise for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and
    d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_neg_base, *args, **kwargs)(v, t)

def prox_neg_entr(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = x\\log(x)` applied
    elementwise for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and
    d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_neg_entr_base, *args, **kwargs)(v, t)

def prox_neg_log(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = -\\log(x)` applied
    elementwise for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and
    d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_neg_log_base, *args, **kwargs)(v, t)

def prox_pos(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = \\max(x,0)` applied
    elementwise for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and
    d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    return prox_scale(prox_pos_base, *args, **kwargs)(v, t)

def prox_abs_base(v, t):
	"""Proximal operator of :math:`f(x) = |x|`.
	"""
	return apply_to_nonzeros(lambda y: np.maximum(y - t, 0) + np.minimum(y + t, 0), v)

def prox_constant_base(v, t):
	"""Proximal operator of :math:`f(x) = c` for any constant :math:`c`.
	"""
	return v

def prox_exp_base(v, t):
	"""Proximal operator of :math:`f(x) = \\exp(x)`.
	"""
	if sparse.issparse(v):
		v = v.todense()
	return v - lambertw(np.exp(v + np.log(t)))

def prox_huber_base(v, t, M):
	"""Proximal operator of
	.. math::
        f(x) =
            \\begin{cases}
                2M|x|-M^2 & \\text{for } |x| \\geq |M| \\\\
                      |x|^2 & \\text{for } |x| \\leq |M|
            \\end{cases}
    applied elementwise, where :math:`M` is a positive scalar.
	"""
	return apply_to_nonzeros(lambda y: np.where(np.abs(y) <= (M + 2*M*t), y / (1 + 2*t), y - 2*M*t*np.sign(y)), v)

def prox_identity_base(v, t):
	"""Proximal operator of :math:`f(x) = x`.
	"""
	if sparse.issparse(v):
		v = v.todense()
	return v - t

def prox_neg_base(v, t):
	"""Proximal operator of :math:`f(x) = -\\min(x,0)`, where the minimum is taken elementwise.
	"""
	return apply_to_nonzeros(lambda y: np.where(y + t <= 0, y + t, np.where(y >= 0, y, 0)), v)

def prox_neg_entr_base(v, t):
	"""Proximal operator of :math:`f(x) = x\\log(x)`.
	"""
	if sparse.issparse(v):
		v = v.todense()
	return t * lambertw(np.exp((v/t - 1) - np.log(t)))

def prox_neg_log_base(v, t):
	"""Proximal operator of :math:`f(x) = -\\log(x)`.
	"""
	if sparse.issparse(v):
		v = v.todense()
	return (v + np.sqrt(v**2 + 4*t)) / 2

def prox_pos_base(v, t):
	"""Proximal operator of :math:`f(x) = \\max(x,0)`, where the maximum is taken elementwise.
	"""
	return apply_to_nonzeros(lambda y: np.where(y - t >= 0, y - t, np.where(y <= 0, y, 0)), v)



import numpy as np
from scipy import sparse
from a2dr.proximal.interface import NUMPY_FUNS, SPARSE_FUNS, apply_to_nonzeros
from a2dr.proximal.projection import proj_l1
from a2dr.proximal.composition import prox_scale

def prox_norm1(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = \\|x\\|_1`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(v):
        v = np.array([v])
    return prox_scale(prox_norm1_base, *args, **kwargs)(v, t)

def prox_norm2(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = \\|x\\|_2`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(v):
        v = np.array([v])
    return prox_scale(prox_norm2_base, *args, **kwargs)(v, t)

def prox_norm_inf(v, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(ax-b) + c^Tx + d\\|x\\|_2^2`, where :math:`f(x) = \\|x\\|_{\\infty}`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(v):
        v = np.array([v])
    return prox_scale(prox_norm_inf_base, *args, **kwargs)(v, t)

def prox_norm_nuc(B, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(aB-b) + cB + d\\|B\\|_F^2`, where :math:`f(B) = \\|B\\|_*`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(B):
        B = np.array([[B]])
    return prox_scale(prox_norm_nuc_base, *args, **kwargs)(B, t)

def prox_norm_fro(B, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(aB-b) + cB + d\\|B\\|_F^2`, where :math:`f(B) = \\|B\\|_2`
    for scalar t > 0, and the optional arguments are a = scale, b = offset, c = lin_term, and d = quad_term.
    We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1, b = 0, c = 0, and d = 0.
    """
    if np.isscalar(B):
        B = np.array([[B]])
    return prox_scale(prox_norm_fro_base, *args, **kwargs)(B, t)

def prox_group_lasso(B, t = 1, *args, **kwargs):
    """Proximal operator of :math:`tf(aB-b) + cB + d\\|B\\|_F^2`, where :math:`f(B) = \\|B\\|_{2,1}` is the
    group lasso of :math:`B`, for scalar t > 0, and the optional arguments are a = scale, b = offset,
    c = lin_term, and d = quad_term. We must have t > 0, a = non-zero, and d >= 0. By default, t = 1, a = 1,
    b = 0, c = 0, and d = 0.
    """
    if np.isscalar(B):
        B = np.array([[B]])
    return prox_scale(prox_group_lasso_base, *args, **kwargs)(B, t)

def prox_norm1_base(v, t):
    """Proximal operator of :math:`f(x) = \\|x\\|_1`.
    """
    return apply_to_nonzeros(lambda y: np.maximum(y - t, 0) - np.maximum(-y - t, 0), v)

def prox_norm2_base(v, t):
    """Proximal operator of :math:`f(x) = \\|x\\|_2`.
    """
    FUNS = SPARSE_FUNS if sparse.issparse(v) else NUMPY_FUNS
    norm, zeros = FUNS["norm"], FUNS["zeros"]

    if np.isscalar(v):
        v_norm = abs(v)
    elif len(v.shape) == 1:
        v_norm = norm(v,2)
    elif len(v.shape) == 2:
        v_norm = norm(v,'fro')

    if v_norm == 0:
        return zeros(v.shape)
    else:
        return np.maximum(1 - t/v_norm, 0) * v

def prox_norm_inf_base(v, t):
    """Proximal operator of :math:`f(x) = \\|x\\|_{\\infty}`.
    """
    return v - t * proj_l1(v/t)

def prox_norm_fro_base(B, t):
    """Proximal operator of :math:`f(B) = \\|B\\|_2`, the Frobenius norm of :math:`B`.
    """
    U, s, Vt = np.linalg.svd(B, full_matrices=False)
    s_new = prox_norm2_base(s, t)
    # s_norm = np.linalg.norm(s, 2)
    # s_new = np.zeros(s.shape) if s_norm == 0 else np.maximum(1 - t/s_norm, 0) * s
    return U.dot(np.diag(s_new)).dot(Vt)

def prox_norm_nuc_base(B, t):
    """Proximal operator of :math:`f(B) = \\|B\\|_*`, the nuclear norm of :math:`B`.
    """
    U, s, Vt = np.linalg.svd(B, full_matrices=False)
    s_new = prox_norm1_base(s, t)
    # s_new = np.maximum(s - t, 0)
    return U.dot(np.diag(s_new)).dot(Vt)

def prox_group_lasso_base(B, t):
    """Proximal operator of :math:`f(B) = \\|B\\|_{2,1} = \\sum_j \\|B_j\\|_2`, the group lasso of :math:`B`,
    where :math:`B_j` is the j-th column of :math:`B`.
    """
    # FUNS = SPARSE_FUNS if sparse.issparse(B) else NUMPY_FUNS
    # vstack, hstack = FUNS["vstack"], FUNS["hstack"]
    # prox_cols = [prox_norm2(B[:, j], t) for j in range(B.shape[1])]
    # return vstack(prox_cols).T if prox_cols[0].ndim == 1 else hstack(prox_cols)
    if sparse.issparse(B):
        B = B.tocsc()
        prox_cols = [prox_norm2(B[:, j], t) for j in range(B.shape[1])]
        return sparse.hstack(prox_cols)
    else:
        prox_cols = [prox_norm2(B[:, j], t) for j in range(B.shape[1])]
        return np.column_stack(prox_cols)
